# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from traceback import print_exc
from typing import Union

from fabric.exceptions import GroupException
from concurrent.futures import TimeoutError

KME_DICTIONARY = {
    "KME0000": {
        "name": "Test exception"
    },
    "KME0003": {
        "instance": TimeoutError,
        "name": "Action took too long to complete and timed out"
    },
    "KME0004": {
        "name": "There are no workers defined in the cluster scheme"
    },
    "KME0005": {
        "name": "{hostnames} are not sudoers"
    },
    "KME0007": {
        "name": "Docker CRI can not be used with endpoints registry definition."
    },
}


# TODO: support for more complex KME00XX objects with custom constructors
class KME(RuntimeError):
    def __init__(self, code, **kwargs):
        self.code = code
        self.kme = KME_DICTIONARY.get(self.code)
        if self.kme is None:
            raise ValueError('An error was raised with an unknown error code')
        self.message = self.kme.get('name').format(**kwargs)
        super().__init__(self.message)

    def __str__(self):
        return self.code + ": " + self.message


class FailException(Exception):
    def __init__(self, message='', reason: Union[str, Exception] = '', hint=''):
        super().__init__(message)
        self.message = message
        self.reason = reason
        self.hint = hint


def pretty_print_error(reason: Union[str, Exception], log=None) -> None:
    """
    Parses the passed error and nicely displays its name and structure depending on what was passed.
    The method outputs to stdout by default, but will use the logger if one is specified.
    :param reason: an object containing an exception or other error (must be able to be represented
    as a string)
    :param log: logger object, if you need to write a log there
    :return: None
    """

    if reason == "":
        return

    if isinstance(reason, KME):
        if log:
            log.critical(reason)
        else:
            sys.stderr.write(str(reason))

        return

    for dictionary_code, dictionary_kme in KME_DICTIONARY.items():
        if dictionary_kme.get('instance') and isinstance(reason, dictionary_kme['instance']):
            kme = KME(dictionary_code)

            if log:
                log.critical(kme)
            else:
                sys.stderr.write(str(kme))

            return

    if isinstance(reason, GroupException):
        description = "KME0002: Remote group exception"

        if log:
            log.critical(description)
        else:
            sys.stderr.write(f"{description}\n")

        for connection, result in reason.result.items():
            if log:
                log.critical("%s:" % connection.host)
            else:
                sys.stderr.write("\n%s:" % connection.host)

            found_dictionary_code = None
            for dictionary_code, dictionary_kme in KME_DICTIONARY.items():
                if dictionary_kme.get('instance') \
                        and isinstance(result, dictionary_kme['instance']):
                    found_dictionary_code = dictionary_code
                    break

            if found_dictionary_code:
                kme = KME(found_dictionary_code)
                if log:
                    log.critical("\t" + str(kme))
                else:
                    sys.stderr.write("\n\t%s\n" % str(kme))
            else:
                if log:
                    log.critical("\t" + str(result).replace("\n", "\n\t"))
                else:
                    sys.stderr.write("\n\t%s\n" % str(result).replace("\n", "\n\t"))

        return

    if isinstance(reason, Exception):
        if log:
            log.critical('KME0001: Unexpected exception', exc_info=reason)
        else:
            sys.stderr.write("KME0001: Unexpected exception\n\n")
            print_exc()
    else:
        if log:
            log.critical(reason)
        else:
            sys.stderr.write(reason + "\n")



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from typing import List, Dict

import fabric

from kubemarine import yum, apt
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.executor import RemoteExecutor
from kubemarine.core.group import NodeGroup, NodeGroupResult
from kubemarine.core.yaml_merger import default_merger


ERROR_GLOBAL_ASSOCIATIONS_REDEFINED_MULTIPLE_OS = \
    "It is not supported to customize services.packages.associations section " \
    "if nodes have different OS families. " \
    "Please move the section to corresponding services.packages.associations.<os_family> section."

ERROR_MULTIPLE_PACKAGE_VERSIONS_DETECTED = \
    "Multiple package versions detected %s for package '%s'. " \
    "Align them to the single version manually or using corresponding task of install procedure. " \
    "Alternatively, specify cache_versions=false for corresponding association."

ERROR_SEMANAGE_NOT_MANAGED_DEBIAN = "semanage is not managed for debian OS family by KubeMarine"


def enrich_inventory_associations(inventory, cluster: KubernetesCluster):
    associations: dict = inventory['services']['packages']['associations']
    enriched_associations = {}

    # Move associations for OS families and merge with globals
    for association_name in get_associations_os_family_keys():
        os_associations: dict = deepcopy(cluster.globals['packages']['common_associations'])
        if association_name == 'debian':
            del os_associations['semanage']
        for association_params in os_associations.values():
            del association_params['groups']
        default_merger.merge(os_associations, associations.pop(association_name))
        enriched_associations[association_name] = os_associations

    # Check remained associations section if they are customized at global level.
    if associations:
        os_family = cluster.get_os_family()
        if os_family == 'multiple':
            raise Exception(ERROR_GLOBAL_ASSOCIATIONS_REDEFINED_MULTIPLE_OS)
        elif os_family not in ('unknown', 'unsupported'):
            # move remained associations properties to the specific OS family section and merge with priority
            default_merger.merge(enriched_associations[os_family], associations)

    if 'semanage' in enriched_associations['debian']:
        raise Exception(ERROR_SEMANAGE_NOT_MANAGED_DEBIAN)

    inventory['services']['packages']['associations'] = enriched_associations

    return inventory


def enrich_inventory_packages(inventory: dict, _):
    for _type in ['install', 'upgrade', 'remove']:
        packages_list = inventory['services']['packages'].get(_type)
        if isinstance(packages_list, list):
            inventory['services']['packages'][_type] = {
                'include': packages_list
            }

    return inventory


def enrich_inventory_include_all(inventory: dict, _):
    for _type in ['upgrade', 'remove']:
        packages: dict = inventory['services']['packages'].get(_type)
        if packages is not None:
            packages.setdefault('include', ['*'])

    return inventory


def cache_package_versions(cluster: KubernetesCluster, inventory: dict, by_initial_nodes=False) -> dict:
    os_ids = cluster.get_os_identifiers()
    different_os = list(set(os_ids.values()))
    if len(different_os) > 1:
        cluster.log.debug(f"Final nodes have different OS families or versions, packages will not be cached. "
                          f"List of (OS family, version): {different_os}")
        return inventory

    os_family = different_os[0][0]
    if os_family in ('unknown', 'unsupported'):
        # For add_node/install procedures we check that OS is supported in prepare.check.system task.
        # For check_iaas procedure it is allowed to have unsupported OS, so skip caching.
        cluster.log.debug("Skip caching of packages for unsupported OS.")
        return inventory

    group = cluster.nodes['all'].get_final_nodes()
    if group.nodes_amount() != group.get_sudo_nodes().nodes_amount():
        # For add_node/install procedures we check that all nodes are sudoers in prepare.check.sudoer task.
        # For check_iaas procedure the nodes might still be not sudoers.
        # Skip caching if any not-sudoer node found.
        cluster.log.debug(f"Some nodes are not sudoers, packages will not be cached.")
        return inventory

    if by_initial_nodes:
        group = group.get_initial_nodes()

    hosts_to_packages = get_all_managed_packages_for_group(group, inventory, by_initial_nodes)
    detected_packages = detect_installed_packages_version_hosts(cluster, hosts_to_packages)

    _cache_package_associations(group, inventory, detected_packages, by_initial_nodes)
    _cache_custom_packages(cluster, inventory, detected_packages, by_initial_nodes)

    cluster.log.debug('Package versions detection finished')
    return inventory


def get_all_managed_packages_for_group(group: NodeGroup, inventory: dict, ensured_association_only: bool = False) \
        -> Dict[str, List[str]]:
    """
    Returns hosts with list of all managed packages for them.
    For associations, only subset of hosts is considered on which the associations are managed by KubeMarine.

    :param group: Group of nodes to get the manager packages for.
    :param inventory: Inventory of the cluster. May be different from the inventory of the cluster instance,
                      if used during finalization.
    :param ensured_association_only: Specify whether to take 'cache_versions' property into account for associations.
                                     Additionally, if true, will skip custom packages.
    :return: List of packages for each relevant host.
    """
    packages_section = inventory['services']['packages']
    hosts_to_packages = {}
    for node in group.get_ordered_members_list():
        os_family = node.get_nodes_os()
        node_associations = packages_section['associations'].get(os_family, {})
        for association_name in node_associations.keys():
            packages = get_association_hosts_to_packages(
                node, inventory, association_name, ensured_association_only)

            packages = next(iter(packages.values()), [])
            hosts_to_packages.setdefault(node.get_host(), []).extend(packages)

    custom_install_packages = inventory['services']['packages'].get('install', {}).get('include', [])
    if not ensured_association_only and custom_install_packages:
        for host in group.get_hosts():
            hosts_to_packages.setdefault(host, []).extend(custom_install_packages)

    return hosts_to_packages


def get_association_hosts_to_packages(group: NodeGroup, inventory: dict, association_name: str,
                                      ensured_association_only: bool = False) \
        -> Dict[str, List[str]]:
    """
    Returns hosts with associated packages list for the specified association name.
    Only subset of hosts is returned on which the association is managed by KubeMarine.

    :param group: Group of nodes to check the applicability of the association.
    :param inventory: Inventory of the cluster. May be different from the inventory of the cluster instance,
                      if used during finalization.
    :param association_name: target association name
    :param ensured_association_only: Specify whether to take 'cache_versions' property into account.
    :return: List of packages for each relevant host.
    """
    cluster = group.cluster

    packages_section = inventory['services']['packages']
    if not packages_section['mandatory'].get(association_name, True):
        return {}

    hosts_to_packages = {}

    if association_name == 'unzip':
        from kubemarine import thirdparties
        relevant_group = thirdparties.get_group_require_unzip(cluster, inventory)
    else:
        groups = cluster.globals['packages']['common_associations'].get(association_name, {}).get('groups', [])
        relevant_group = cluster.create_group_from_groups_nodes_names(groups, [])

    if association_name in ('docker', 'containerd') \
            and association_name != inventory['services']['cri']['containerRuntime']:
        relevant_group = cluster.make_group([])

    relevant_group = relevant_group.intersection_group(group)

    global_cache_versions = packages_section['cache_versions']
    for node in relevant_group.get_ordered_members_list():
        os_family = node.get_nodes_os()
        package_associations = packages_section['associations'].get(os_family, {}).get(association_name, {})
        packages = package_associations.get('package_name', [])

        if isinstance(packages, str):
            packages = [packages]

        if ensured_association_only and not (global_cache_versions and package_associations.get('cache_versions', True)):
            packages = []

        if packages:
            hosts_to_packages[node.get_host()] = packages

    return hosts_to_packages


def _cache_package_associations(group: NodeGroup, inventory: dict,
                                detected_packages: Dict[str, Dict[str, List]], ensured_association_only: bool):
    cluster = group.cluster
    associations = inventory['services']['packages']['associations'][cluster.get_os_family()]
    for association_name, associated_params in associations.items():
        hosts_to_packages = get_association_hosts_to_packages(
            group, inventory, association_name, ensured_association_only)
        if not hosts_to_packages:
            continue

        # Since all nodes have the same OS family in this case,
        # the packages list is the same for all relevant hosts, so take any available.
        packages_list = next(iter(hosts_to_packages.values()))

        final_packages_list = []
        for package in packages_list:
            final_package = _detect_final_package(cluster, detected_packages, package, ensured_association_only)
            final_packages_list.append(final_package)

        # if non-multiple value, then convert to simple string
        # packages can contain multiple package values, like docker package
        # (it has docker-ce, docker-cli and containerd.io packages for installation)
        if len(final_packages_list) == 1:
            final_packages_list = final_packages_list[0]

        associated_params['package_name'] = final_packages_list


def _cache_custom_packages(cluster: KubernetesCluster, inventory: dict,
                           detected_packages: Dict[str, Dict[str, List]], ensured_association_only: bool):
    if ensured_association_only:
        return
    # packages from direct installation section
    custom_install_packages = inventory['services']['packages'].get('install', {})
    if custom_install_packages.get('include', []):
        final_packages_list = []
        for package in custom_install_packages['include']:
            final_package = _detect_final_package(cluster, detected_packages, package, False)
            final_packages_list.append(final_package)
        custom_install_packages['include'] = final_packages_list


def _detect_final_package(cluster: KubernetesCluster, detected_packages: Dict[str, Dict[str, List]],
                          package: str, ensured_association_only: bool) -> str:
    # add package version to list only if it was found as installed
    detected_package_versions = list(filter(lambda version: "not installed" not in version,
                                            detected_packages[package].keys()))

    # if there no versions detected, then return default package from inventory
    if not detected_package_versions:
        return package
    elif len(detected_package_versions) > 1:
        if ensured_association_only:
            raise Exception(ERROR_MULTIPLE_PACKAGE_VERSIONS_DETECTED % (str(detected_packages[package]), package))
        else:
            cluster.log.warning(
                f"Multiple package versions detected {detected_packages[package]} for package '{package}'. "
                f"Use default package '{package}' from inventory.")
            # return default package from inventory if multiple versions detected
            return package
    else:
        return detected_package_versions[0]


def remove_unused_os_family_associations(cluster: KubernetesCluster, inventory: dict):
    final_nodes = cluster.nodes['all'].get_final_nodes()
    for os_family in get_associations_os_family_keys():
        # Do not remove OS family associations section in finalized inventory if any node has this OS family.
        if final_nodes.get_subgroup_with_os(os_family).is_empty():
            del inventory['services']['packages']['associations'][os_family]

    return inventory


def get_associations_os_family_keys():
    return {'debian', 'rhel', 'rhel8'}


def get_package_manager(group: NodeGroup) -> apt or yum:
    os_family = group.get_nodes_os()

    if os_family in ['rhel', 'rhel8']:
        return yum
    elif os_family == 'debian':
        return apt

    raise Exception('Failed to return package manager for unknown or multiple OS')


def ls_repofiles(group: NodeGroup, **kwargs) -> NodeGroupResult:
    return get_package_manager(group).ls_repofiles(group, **kwargs)


def backup_repo(group: NodeGroup, repo_filename="*", **kwargs) -> NodeGroupResult:
    return get_package_manager(group).backup_repo(group, repo_filename, **kwargs)


def add_repo(group: NodeGroup, repo_data="", repo_filename="predefined", **kwargs) -> NodeGroupResult:
    return get_package_manager(group).add_repo(group, repo_data, repo_filename, **kwargs)


def clean(group: NodeGroup, mode="all", **kwargs) -> NodeGroupResult:
    return get_package_manager(group).clean(group, mode, **kwargs)


def install(group: NodeGroup, include=None, exclude=None, **kwargs) -> NodeGroupResult:
    return get_package_manager(group).install(group, include, exclude, **kwargs)


def remove(group: NodeGroup, include=None, exclude=None, **kwargs) -> NodeGroupResult:
    return get_package_manager(group).remove(group, include, exclude, **kwargs)


def upgrade(group: NodeGroup, include=None, exclude=None, **kwargs) -> NodeGroupResult:
    return get_package_manager(group).upgrade(group, include, exclude, **kwargs)


def no_changes_found(group: NodeGroup, action: callable, result: fabric.runners.Result) -> bool:
    pkg_mgr = get_package_manager(group)
    if action is install:
        action = pkg_mgr.install
    elif action is upgrade:
        action = pkg_mgr.upgrade
    elif action is remove:
        action = pkg_mgr.remove
    return pkg_mgr.no_changes_found(action, result)


def get_detect_package_version_cmd(os_family: str, package_name: str) -> str:
    if os_family in ["rhel", "rhel8"]:
        cmd = r"rpm -q %s" % package_name
    else:
        cmd = r"dpkg-query -f '${Package}=${Version}\n' -W %s" % package_name

    # This is WA for RemoteExecutor, since any package failed others are not checked
    # TODO: get rid of this WA and use warn=True in sudo
    cmd += ' || true'
    return cmd


def _detect_installed_package_version(group: NodeGroup, package: str) -> NodeGroupResult:
    """
    Detect package versions for each host on remote group
    :param group: Group of nodes, where package should be found
    :param package: package name, which version should be detected (eg. 'podman' and 'containerd')
    :return: NodeGroupResults with package version on each host

    Method generates different package query for different OS.

    Note: for Ubuntu/Debian some packages returns multiline results for some queries
    (for example docker-ce* returns docker-ce and docker-ce-cli).
    """

    os_family = group.get_nodes_os()
    package_name = get_package_name(os_family, package)

    cmd = get_detect_package_version_cmd(os_family, package_name)
    return group.sudo(cmd)


def _parse_node_detected_package(result: fabric.runners.Result, package: str) -> str:
    node_detected_package = result.stdout.strip() + result.stderr.strip()
    # consider version, which ended with special symbol = or - as not installed
    # (it is possible in some cases to receive "containerd=" version)
    if "not installed" in node_detected_package or "no packages found" in node_detected_package \
            or node_detected_package[-1] == '=' or node_detected_package[-1] == '-':
        node_detected_package = f"not installed {package}"

    return node_detected_package


def detect_installed_packages_version_hosts(cluster: KubernetesCluster, hosts_to_packages: Dict[str, List[str]]) \
        -> Dict[str, Dict[str, List]]:
    """
    Detect grouped packages versions for specified list of packages for each remote host.

    :param cluster: KubernetesCluster instance
    :param hosts_to_packages: Remote hosts with list of packages to detect versions.
    :return: Dictionary with grouped versions for each queried package, pointing to list of hosts,
        e.g. {"foo" -> {"foo-1": [host1, host2]}, "bar" -> {"bar-1": [host1], "bar-2": [host2]}}
    """
    for host, packages_list in hosts_to_packages.items():
        if isinstance(packages_list, str):
            packages_list = [packages_list]
        # deduplicate
        hosts_to_packages[host] = list(set(packages_list))

    with RemoteExecutor(cluster) as exe:
        for host, packages_list in hosts_to_packages.items():
            node = cluster.make_group([host])
            for package in packages_list:
                _detect_installed_package_version(node, package)

    raw_result = exe.get_last_results()
    if not raw_result:
        return {}

    results: Dict[str, Dict[str, List]] = {}

    for conn, multiple_results in raw_result.items():
        multiple_results = list(multiple_results.values())
        host = conn.host
        packages_list = hosts_to_packages[host]
        for i, package in enumerate(packages_list):
            node_detected_package = _parse_node_detected_package(multiple_results[i], package)
            results.setdefault(package, {}).setdefault(node_detected_package, []).append(host)

    return results


def get_package_name(os_family: str, package: str) -> str:
    """
    Return the pure package name, without any part of version
    """

    import re

    package_name = ""
    
    if package:
        if os_family in ["rhel", "rhel8"]:
            # regexp is needed to split package and its version, the pattern start with '-' then should be number or '*'
            package_name = re.split(r'-[\d,\*]', package)[0]
        else:
            # in ubuntu it is much easier to parse package name
            package_name = package.split("=")[0]

    return package_name


def search_package(group: NodeGroup, package: str, **kwargs) -> NodeGroupResult:
    return get_package_manager(group).search(group, package, **kwargs)


def create_repo_file(group: NodeGroup, repo_data, repo_file):
    get_package_manager(group).create_repo_file(group, repo_data, repo_file)


def get_repo_filename(group: NodeGroup, repo_filename="predefined"):
    return get_package_manager(group).get_repo_file_name(repo_filename)


# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod

from kubemarine.core.action import Action


class Patch(ABC):
    def __init__(self, identifier: str):
        self.identifier = identifier

    @property
    @abstractmethod
    def action(self) -> Action:
        pass

    @property
    @abstractmethod
    def description(self) -> str:
        pass



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import StringIO

import toml
import yaml
import json

from distutils.util import strtobool
from kubemarine import system, packages
from kubemarine.core import utils
from kubemarine.core.executor import RemoteExecutor
from kubemarine.core.group import NodeGroup


def install(group: NodeGroup):
    with RemoteExecutor(group.cluster) as exe:
        for node in group.get_ordered_members_list(provide_node_configs=True):
            os_specific_associations = group.cluster.get_associations_for_node(node['connect_to'], 'containerd')

            group.cluster.log.debug("Installing latest containerd and podman on %s node" % node['name'])
            # always install latest available containerd and podman
            packages.install(node['connection'], include=os_specific_associations['package_name'])

            # remove previous config.toml to avoid problems in case when previous config was broken
            node['connection'].sudo("rm -f %s && sudo systemctl restart %s"
                                    % (os_specific_associations['config_location'],
                                       os_specific_associations['service_name']))

            system.enable_service(node['connection'], os_specific_associations['service_name'], now=True)
    return exe.get_last_results_str()


def configure(group: NodeGroup):
    log = group.cluster.log

    log.debug("Uploading crictl configuration for containerd...")
    crictl_config = yaml.dump({"runtime-endpoint": "unix:///run/containerd/containerd.sock"})
    utils.dump_file(group.cluster, crictl_config, 'crictl.yaml')
    group.put(StringIO(crictl_config), '/etc/crictl.yaml', backup=True, sudo=True)

    config_string = ""
    # double loop is used to make sure that no "simple" `key: value` pairs are accidentally assigned to sections
    containerd_config = group.cluster.inventory["services"]["cri"]['containerdConfig']
    runc_options_path = 'plugins."io.containerd.grpc.v1.cri".containerd.runtimes.runc.options'
    if not isinstance(containerd_config[runc_options_path]['SystemdCgroup'], bool):
        containerd_config[runc_options_path]['SystemdCgroup'] = \
            bool(strtobool(containerd_config[runc_options_path]['SystemdCgroup']))
    for key, value in containerd_config.items():
        # first we process all "simple" `key: value` pairs
        if not isinstance(value, dict):
            config_string += f"{toml.dumps({key: value})}"
    for key, value in containerd_config.items():
        # next we process all "complex" `key: dict_value` pairs, representing named sections
        if isinstance(value, dict):
            config_string += f"\n[{key}]\n{toml.dumps(value)}"

    # if there are any insecure registries in containerd config, then it is required to configure them for podman too
    config_toml = toml.loads(config_string)
    insecure_registries = []
    registry = config_toml.get('plugins', {}).get('io.containerd.grpc.v1.cri', {}).get('registry', {})
    if registry.get('mirrors'):
        for mirror, mirror_conf in registry['mirrors'].items():
            is_insecure = registry.get('configs', {}).get(mirror, {}).get('tls', {}).get('insecure_skip_verify', False)
            for endpoint in mirror_conf.get('endpoint', []):
                if "http://" in endpoint:
                    is_insecure = True
                    break
            if is_insecure:
                insecure_registries.append(mirror)
    # save 'auth.json' if there are credentials for registry
    auth_registries = {"auths": {}}
    if registry.get('configs'):
        registry_configs = registry['configs']
        for auth_registry in registry_configs:
            auth_registries['auths'][auth_registry] = {}
            if registry_configs[auth_registry].get('auth', {}).get('auth', ''):
                auth_registries['auths'][auth_registry]['auth'] = registry_configs[auth_registry]['auth']['auth']
        auth_json = json.dumps(auth_registries)
        group.put(StringIO(auth_json), "/etc/containers/auth.json", backup=True, sudo=True)
        group.sudo("chmod 600 /etc/containers/auth.json")
    if insecure_registries:
        log.debug("Uploading podman configuration...")
        podman_registries = f"[registries.insecure]\nregistries = {insecure_registries}\n"
        utils.dump_file(group.cluster, podman_registries, 'podman_registries.conf')
        group.sudo("mkdir -p /etc/containers/")
        group.put(StringIO(podman_registries), "/etc/containers/registries.conf", backup=True, sudo=True)
    else:
        log.debug("Removing old podman configuration...")
        group.sudo("rm -f /etc/containers/registries.conf")

    utils.dump_file(group.cluster, config_string, 'containerd-config.toml')
    with RemoteExecutor(group.cluster) as exe:
        for node in group.get_ordered_members_list(provide_node_configs=True):
            os_specific_associations = group.cluster.get_associations_for_node(node['connect_to'], 'containerd')
            log.debug("Uploading containerd configuration to %s node..." % node['name'])
            node['connection'].put(StringIO(config_string), os_specific_associations['config_location'], backup=True,
                                   sudo=True, mkdir=True)
            log.debug("Restarting Containerd on %s node..." % node['name'])
            node['connection'].sudo(f"chmod 600 {os_specific_associations['config_location']} && "
                                    f"sudo systemctl restart {os_specific_associations['service_name']} && "
                                    f"systemctl status {os_specific_associations['service_name']}")
    return exe.get_last_results_str()


def prune(group):
    return group.sudo('crictl rm -fa; '
                      'sudo crictl rmp -fa; '
                      'sudo crictl rmi -a; '
                      'sudo ctr content ls -q | xargs -r sudo ctr content rm', warn=True)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser
import io

import fabric

from kubemarine.core import utils
from kubemarine.core.group import NodeGroupResult, NodeGroup


def ls_repofiles(group, **kwargs):
    return group.sudo('ls -la /etc/yum.repos.d', **kwargs)


def backup_repo(group, repo_filename="*", **kwargs):
    if not group.cluster.inventory['services']['packages']['package_manager']['replace-repositories']:
        group.cluster.log.debug("Skipped - repos replacement disabled in configuration")
        return
    # all files in directory will be renamed: xxx.repo -> xxx.repo.bak
    # if there already any files with ".bak" extension, they should not be renamed to ".bak.bak"!
    return group.sudo("find /etc/yum.repos.d/ -type f -name '%s.repo' | "
                      "sudo xargs -t -iNAME mv -bf NAME NAME.bak" % repo_filename, **kwargs)


def add_repo(group, repo_data="", repo_filename="predefined", **kwargs):
    create_repo_file(group, repo_data, get_repo_file_name(repo_filename))
    return group.sudo('yum clean all && sudo yum updateinfo', **kwargs)


def get_repo_file_name(repo_filename="predefined"):
    return '/etc/yum.repos.d/%s.repo' % repo_filename


def create_repo_file(group, repo_data, repo_file):
    # if repo_data is dict, then convert it to string with config inside
    if isinstance(repo_data, dict):
        config = configparser.ConfigParser()
        for repo_id, data in repo_data.items():
            config[repo_id] = data
        repo_data = io.StringIO()
        config.write(repo_data)
    else:
        repo_data = io.StringIO(utils.read_external(repo_data))
    group.put(repo_data, repo_file, sudo=True)


def clean(group, mode="all", **kwargs):
    return group.sudo("yum clean %s" % mode, **kwargs)


def get_install_cmd(include: str or list, exclude=None) -> str:
    if isinstance(include, list):
        include = ' '.join(include)
    command = 'yum install -y %s' % include

    if exclude is not None:
        if isinstance(exclude, list):
            exclude = ','.join(exclude)
        command += ' --exclude=%s' % exclude
    command += f"; rpm -q {include}; if [ $? != 0 ]; then echo \"Failed to check version for some packages. " \
               f"Make sure packages are not already installed with higher versions. " \
               f"Also, make sure user-defined packages have rpm-compatible names. \"; exit 1; fi "

    return command


def install(group, include=None, exclude=None, **kwargs):
    if include is None:
        raise Exception('You must specify included packages to install')

    command = get_install_cmd(include, exclude)

    return group.sudo(command, **kwargs)


def remove(group, include=None, exclude=None, **kwargs):
    if include is None:
        raise Exception('You must specify included packages to remove')

    if isinstance(include, list):
        include = ' '.join(include)
    command = 'yum remove -y %s' % include

    if exclude is not None:
        if isinstance(exclude, list):
            exclude = ','.join(exclude)
        command += ' --exclude=%s' % exclude

    return group.sudo(command, **kwargs)


def upgrade(group, include=None, exclude=None, **kwargs):
    if include is None:
        raise Exception('You must specify included packages to upgrade')

    if isinstance(include, list):
        include = ' '.join(include)
    command = 'yum upgrade -y %s' % include

    if exclude is not None:
        if isinstance(exclude, list):
            exclude = ','.join(exclude)
        command += ' --exclude=%s' % exclude

    return group.sudo(command, **kwargs)


def no_changes_found(action: callable, result: fabric.runners.Result) -> bool:
    if action is install:
        return "Nothing to do" in result.stdout
    elif action is upgrade:
        return "No packages marked for update" in result.stdout
    elif action is remove:
        return "No Packages marked for removal" in result.stdout
    else:
        raise Exception(f"Unknown action {action}")


def search(group: NodeGroup, package: str, **kwargs) -> NodeGroupResult:
    if package is None:
        raise Exception('You must specify package to search')
    command = 'yum list %s || echo "Package is unavailable"' % package
    return group.sudo(command, **kwargs)


# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys
from pygelf import gelf, GelfTcpHandler, GelfUdpHandler, GelfTlsHandler, GelfHttpHandler

from copy import deepcopy
from typing import List, Optional

VERBOSE = 5
gelf.LEVELS.update({VERBOSE: 8})

DEFAULT_FORMAT = '%(asctime)s %(name)s %(levelname)s %(message)s'

BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)

RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"

COLORS = {
    'RESET': RESET_SEQ,
    'BOLD': BOLD_SEQ,
    'BLACK': COLOR_SEQ % (30 + BLACK),
    'RED': COLOR_SEQ % (30 + RED),
    'GREEN': COLOR_SEQ % (30 + GREEN),
    'YELLOW': COLOR_SEQ % (30 + YELLOW),
    'BLUE': COLOR_SEQ % (30 + BLUE),
    'MAGENTA': COLOR_SEQ % (30 + MAGENTA),
    'CYAN': COLOR_SEQ % (30 + CYAN),
    'WHITE': COLOR_SEQ % (30 + WHITE),
}

COLORS_SCHEME = {
    'WARNING': 'YELLOW',
    'VERBOSE': 'BLUE',
    'INFO': 'GREEN',
    'ERROR': 'RED',
    'CRITICAL': 'RED'
}

LOGGING_LEVELS_BY_NAME = {
    '5': VERBOSE,
    'verbose': VERBOSE,
    '10': logging.DEBUG,
    'debug': logging.DEBUG,
    '20': logging.INFO,
    'info': logging.INFO,
    '30': logging.WARNING,
    'warn': logging.WARNING,
    'warning': logging.WARNING,
    '40': logging.ERROR,
    'error': logging.ERROR,
    '50': logging.CRITICAL,
    'critical': logging.CRITICAL
}

LOGGING_NAMES_BY_LEVEL = {
    VERBOSE: 'verbose',
    logging.DEBUG: 'debug',
    logging.INFO: 'info',
    logging.WARNING: 'warning',
    logging.ERROR: 'error',
    logging.CRITICAL: 'critical'
}


class EnhancedLogger(logging.Logger):
    def __init__(self, name, level=logging.NOTSET):
        super().__init__(name, level)
        logging.addLevelName(VERBOSE, 'VERBOSE')

    def verbose(self, msg, *args, **kwargs):
        if self.isEnabledFor(VERBOSE):
            self._log(VERBOSE, msg, args, **kwargs)


logging.setLoggerClass(EnhancedLogger)


class LogFormatter(logging.Formatter):
    def __init__(self, fmt=None, datefmt=None, style='%', colorize=False, correct_newlines=False):
        super().__init__(fmt, datefmt, style)
        self.colorize = colorize
        self.correct_newlines = correct_newlines

    def _format(self, record):
        s = super().format(record)
        if self.colorize and record.levelname in COLORS_SCHEME:
            s = '$__COLOR_' + COLORS_SCHEME[record.levelname] + s + '$__COLOR_RESET'
        for color_name, color_code in COLORS.items():
            if self.colorize:
                s = s.replace('$__COLOR_' + color_name, color_code)
            else:
                s = s.replace('$__COLOR_' + color_name, '')
        return s

    def format(self, record):
        messages = str(record.msg).split('\n')
        if self.correct_newlines and len(messages):
            subrecord = logging.makeLogRecord(record.__dict__)
            s = ''
            for message in messages:
                if s != '':
                    s += '\n'
                subrecord.msg = message
                s += self._format(subrecord)
            return s
        else:
            return self._format(record)


class StdoutHandler(logging.StreamHandler):
    def __init__(self):
        super().__init__(sys.stdout)

    def emit(self, record):
        if hasattr(record, 'ignore_stdout') and getattr(record, 'ignore_stdout'):
            return
        super().emit(record)
        # TODO: if output stuck, then add here self.flush()
        # More about, see at https://stackoverflow.com/questions/16633911


class FileHandlerWithHeader(logging.FileHandler):
    def __init__(self, filename, header=None, mode='a', encoding=None, delay=0):
        # Store the header information.
        self.header = header

        # Determine if the file pre-exists
        self.file_pre_exists = os.path.exists(filename)

        # Call the parent __init__
        logging.FileHandler.__init__(self, filename, mode, encoding, delay)

        # Write the header if delay is False and a file stream was created.
        if not delay and header and self.stream is not None:
            self.stream.write('%s\n' % header)

    def emit(self, record):
        # Create the file stream if not already created.
        if self.stream is None:
            self.stream = self._open()

            # If the file pre_exists, it should already have a header.
            # Else write the header to the file so that it is the first line.
            if not self.file_pre_exists:
                self.stream.write('')
                if self.header:
                    self.stream.write('%s\n' % self.header)

        # Call the parent class emit function.
        logging.FileHandler.emit(self, record)


class LogHandler:

    def __init__(self,
                 target: str,
                 level: str,
                 colorize: bool = False,
                 correct_newlines: bool = False,
                 filemode: str = 'a',
                 format: str = DEFAULT_FORMAT,
                 datefmt: str = None,
                 header: str = None,
                 **kwargs):

        self._colorize = colorize
        self._correct_newlines = correct_newlines
        self._format = format
        self._datefmt = datefmt
        self._header = header

        self._formatter = LogFormatter(self._format, self._datefmt,
                                       colorize=self._colorize,
                                       correct_newlines=self._correct_newlines)

        if target.lower() == 'stdout':
            self._target = 'stdout'
            self.handler = StdoutHandler()
        elif target.lower() == 'graylog':
            self._target = 'graylog'
            if not kwargs.get('host'):
                raise Exception('Graylog host is not defined')
            if not kwargs.get('port'):
                raise Exception(f'Graylog port is not defined for "{kwargs["host"]}"')
            if not kwargs.get('type'):
                raise Exception(f'Graylog type is not defined for "{kwargs["host"]}:{kwargs["port"]}"')
            handler_options = {
                'host': kwargs['host'],
                'port': kwargs['port'],
                '_app_name': kwargs.get('appname', 'kubemarine'),
                'debug': kwargs.get('debug', False),
                'version': kwargs.get('version', '1.1')
            }
            if kwargs['type'] == 'tcp':
                self.handler = GelfTcpHandler(**handler_options)
            elif kwargs['type'] == 'udp':
                handler_options['compress'] = kwargs.get('compress', True)
                handler_options['chunk_size'] = kwargs.get('chunk_size', 1300)
                self.handler = GelfUdpHandler(**handler_options)
            elif kwargs['type'] == 'tls':
                handler_options['validate'] = kwargs.get('validate', True)
                handler_options['ca_certs'] = kwargs.get('ca_certs')
                handler_options['certfile'] = kwargs.get('certfile')
                handler_options['keyfile'] = kwargs.get('keyfile')
                self.handler = GelfTlsHandler(**handler_options)
            elif kwargs['type'] == 'http':
                handler_options['compress'] = kwargs.get('compress', True)
                handler_options['path'] = kwargs.get('path', '/gelf')
                handler_options['timeout'] = kwargs.get('timeout', 5)
                self.handler = GelfHttpHandler(**handler_options)
            else:
                raise Exception(f'Unknown Graylog type "{kwargs["type"]}" for "{kwargs["host"]}:{kwargs["port"]}"')
        else:
            self._target = target
            # Output produced by remote commands might contain characters which cannot be encoded on Windows deployer.
            # Specify explicitly utf-8 encoding which is native to the remote machines.
            self.handler = FileHandlerWithHeader(self._target, mode=filemode, header=self._header, encoding='utf-8')

        self._level = LOGGING_LEVELS_BY_NAME.get(level)
        if self._level is None:
            raise Exception(f'Failed to create logger - unknown logging level: "{level}"')
        self.handler.setLevel(self._level)

        self.handler.setFormatter(self._formatter)

    def __str__(self):
        return f'target: {self._target}, level: {LOGGING_NAMES_BY_LEVEL[self._level]}, colorize: {self._colorize}, datefmt: {self._datefmt}, format: {self._format}'

    def append_to_logger(self, logger) -> None:
        logger.addHandler(self.handler)

    def has_stdout_target(self) -> bool:
        return self._target == 'stdout'


class Log:

    def __init__(self, raw_inventory, handlers: List[LogHandler]):
        self._logger = logging.getLogger(raw_inventory.get('cluster_name', 'cluster.local'))
        self._logger.setLevel(VERBOSE)

        if self._logger.hasHandlers():
            self._logger.handlers.clear()

        for handler in handlers:
            handler.append_to_logger(self._logger)

    @property
    def logger(self) -> EnhancedLogger:
        return self._logger


def parse_log_argument(argument: str) -> LogHandler:
    """
    Parse raw CLI arguments and verify for required parameters
    :param argument: Raw CLI argument string. For example: test.log;level=verbose;colorize=true
    :return: Initialized LogHandler
    """
    parameters = {}
    argument_parts = argument.split(';')
    if not argument_parts:
        raise Exception('Defined logger do not contain parameters')
    parameters['target'] = argument_parts[0]
    for parameter in argument_parts[1:]:
        if parameter == '':
            continue
        key, value, *rest = parameter.split('=')
        if key in ['colorize', 'correct_newlines', 'debug', 'compress', 'validate']:
            value = value.lower() in ['true', '1']
        elif key in ['chunk_size', 'timeout', 'port']:
            value = int(value)
        parameters[key] = value
    if not parameters.get('level'):
        raise Exception(f'Logging level is not set for logger "{parameters["target"]}"')
    return LogHandler(**parameters)


def get_dump_debug_filepath(context: dict) -> Optional[str]:
    args = context['execution_arguments']
    if args.get('disable_dump', True):
        return None

    return os.path.join(args['dump_location'], 'debug.log')


def init_log_from_context_args(globals, context, raw_inventory) -> Log:
    """
    Create Log from raw CLI arguments in Cluster context
    :param globals: parsed globals collection
    :param context: context holding execution arguments.
    :param raw_inventory: parsed but not yet enriched inventory
    :return: Initialized Log, based on all parsed logging arguments
    """

    handlers = []
    stdout_specified = False

    args = context['execution_arguments']
    if args.get('log') is not None:
        for argument in args.get('log'):
            handler = parse_log_argument(argument[0])
            if handler.has_stdout_target():
                if stdout_specified:
                    raise Exception('Multiple stdout logs specified')
                else:
                    stdout_specified = True
            handlers.append(handler)

    debug_filepath = get_dump_debug_filepath(context)
    if debug_filepath:
        handlers.append(LogHandler(target=debug_filepath,
                                   **globals['logging']['default_targets']['dump']))

    if not stdout_specified:
        stdout_settings = deepcopy(globals['logging']['default_targets']['stdout'])
        # Globals lacks of colorize property, so calculated value for Windows is "false".
        # But it is still convenient to specify the value explicitly even for Windows for debugging purpose.
        if 'colorize' not in stdout_settings:
            stdout_settings['colorize'] = (os.name != 'nt')
        handlers.append(LogHandler(target='stdout', **stdout_settings))

    log = Log(raw_inventory, handlers)

    log.logger.verbose('Using the following loggers: \n\t%s' % "\n\t".join("- " + str(x) for x in handlers))

    return log



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import json
import os
import shutil
import sys
import time
import tarfile

from typing import Union, Tuple

import yaml
from copy import deepcopy
from datetime import datetime
from collections import OrderedDict

from kubemarine.core.executor import RemoteExecutor
from kubemarine.core.errors import pretty_print_error


def do_fail(message='', reason: Union[str, Exception] = '', hint='', log=None):

    if log:
        log.critical('FAILURE!')
        if message != "":
            log.critical(message)
    else:
        sys.stderr.write("\033[91mFAILURE!")
        if message != "":
            sys.stderr.write(" - " + message + "\n")

    pretty_print_error(reason, log)

    sys.stderr.write("\n")

    # Please do not rewrite this to logging approach:
    # hint should be visible only in stdout and without special formatting
    if hint != "":
        sys.stderr.write(hint)

    if not log:
        sys.stderr.write("\033[0m\n")

    sys.exit(1)


def get_elapsed_string(start, end):
    elapsed = end - start
    hours, remainder = divmod(elapsed, 3600)
    minutes, seconds = divmod(remainder, 60)
    return '{:02}h {:02}m {:02}s'.format(int(hours), int(minutes), int(seconds))


def prepare_dump_directory(location, reset_directory=True):
    if reset_directory and os.path.exists(location) and os.path.isdir(location):
        shutil.rmtree(location)
    os.makedirs(location, exist_ok=True)


def make_ansible_inventory(location, cluster):

    inventory = get_final_inventory(cluster)
    roles = []
    for node in inventory['nodes']:
        for role in node['roles']:
            if role not in roles:
                roles.append(role)

    config = {
        'all': [
            'localhost ansible_connection=local'
        ],
        'cluster:children': []
    }

    already_global_defined = []

    for role in roles:
        config[role] = []
        config['cluster:children'].append(role)
        for node in cluster.nodes[role].get_final_nodes().get_ordered_members_list(provide_node_configs=True):
            record = "%s ansible_host=%s ansible_ssh_user=%s ansible_ssh_private_key_file=%s ip=%s" % \
                     (node['name'],
                      node['connect_to'],
                      node.get('username', cluster.globals['connection']['defaults']['username']),
                      node['keyfile'],
                      node['internal_address'])
            if node.get('address') is not None:
                record += ' external_ip=%s' % node['address']

            if node['name'] not in already_global_defined:
                config['all'].append(record)
                # to avoid duplicate definition in global section we have to check is that was already defined?
                already_global_defined.append(node['name'])

            config[role].append(node['name'])

    config['cluster:vars'] = [
        'ansible_become=True'
    ]

    for group in ['services', 'plugins']:
        if inventory.get(group) is not None:
            for service_name, service_configs in inventory[group].items():
                # write to inventory only plugins, which will be installed
                if group != 'plugins' or service_configs.get('install', False):

                    config['cluster:vars'].append('\n# %s.%s' % (group, service_name))

                    if isinstance(service_configs, dict):

                        if service_configs.get('installation') is not None:
                            del service_configs['installation']
                        if service_configs.get('install') is not None:
                            del service_configs['install']

                        for config_name, config_value in service_configs.items():
                            if isinstance(config_value, dict) or isinstance(config_value, list):
                                config_value = json.dumps(config_value)
                            config['cluster:vars'].append('%s_%s=%s' % (
                                # TODO: Rewrite replace using regex
                                service_name.replace('-', '_').replace('.', '_').replace('/', '_'),
                                config_name.replace('-', '_').replace('.', '_').replace('/', '_'),
                                config_value))
                    else:
                        config_value = json.dumps(service_configs)
                        config['cluster:vars'].append('%s=%s' % (
                            service_name.replace('-', '_').replace('.', '_'),
                            config_value))

    config_compiled = ''
    for section_name, strings in config.items():
        config_compiled += '[%s]' % section_name
        for string in strings:
            config_compiled += '\n' + string
        config_compiled += '\n\n'

    with open_external(location, 'w') as configfile:
        configfile.write(config_compiled)


def get_current_timestamp_formatted():
    return datetime.now().strftime("%Y%m%d-%H%M%S")


def get_final_inventory(cluster, initial_inventory=None):
    if initial_inventory is None:
        inventory = deepcopy(cluster.inventory)
    else:
        inventory = deepcopy(initial_inventory)

    from kubemarine import admission
    from kubemarine.plugins import nginx_ingress
    from kubemarine.procedures import add_node, remove_node, upgrade, migrate_cri

    inventory_finalize_functions = {
        add_node.add_node_finalize_inventory,
        remove_node.remove_node_finalize_inventory,
        upgrade.upgrade_finalize_inventory,
        admission.finalize_inventory,
        nginx_ingress.finalize_inventory,
        migrate_cri.migrate_cri_finalize_inventory
    }

    for finalize_fn in inventory_finalize_functions:
        inventory = finalize_fn(cluster, inventory)

    return inventory


def merge_vrrp_ips(procedure_inventory, inventory):
    if "vrrp_ips" in inventory and len(inventory["vrrp_ips"]) > 0:
        raise Exception("vrrp_ips section already defined, merging not supported yet")
    else:
        inventory["vrrp_ips"] = procedure_inventory["vrrp_ips"]

    if isinstance(inventory, OrderedDict):
        inventory.move_to_end("vrrp_ips", last=False)


def dump_file(context, data, filename):
    if not isinstance(context, dict):
        # cluster is passed instead of the context directly
        cluster = context
        context = cluster.context

    if isinstance(data, io.StringIO):
        data = data.getvalue()
    if isinstance(data, io.TextIOWrapper):
        data = data.read()

    args = context['execution_arguments']
    if not args.get('disable_dump', True) \
            or (filename in ClusterStorage.PRESERVED_DUMP_FILES and context['preserve_inventory']):

        prepare_dump_directory(args.get('dump_location'), reset_directory=False)
        with open_utf8(get_dump_filepath(context, filename), 'w') as file:
            file.write(data)


def get_dump_filepath(context, filename):
    if context.get("dump_filename_prefix"):
        filename = f"{context['dump_filename_prefix']}_{filename}"

    return get_external_resource_path(os.path.join(context['execution_arguments']['dump_location'], filename))


def wait_command_successful(group, command, retries=15, timeout=5, warn=True, hide=False):
    log = group.cluster.log

    while retries > 0:
        log.debug("Waiting for command to succeed, %s retries left" % retries)
        result = group.sudo(command, warn=warn, hide=hide)
        exit_code = list(result.values())[0].exited
        if exit_code == 0:
            log.debug("Command succeeded")
            return
        retries = retries - 1
        time.sleep(timeout)
    raise Exception("Command failed")


def open_utf8(path: str, mode='r'):
    return open(path, mode + 't', encoding='utf-8')


def open_internal(path: str, mode='r'):
    return open_utf8(get_internal_resource_path(path), mode)


def open_external(path: str, mode='r'):
    return open_utf8(get_external_resource_path(path), mode)


def read_internal(path: str) -> str:
    with open_internal(path) as f:
        return f.read()


def read_external(path: str) -> str:
    with open_external(path) as f:
        return f.read()


def get_external_resource_path(path):
    return os.path.abspath(path)


def get_internal_resource_path(path: str) -> str:
    return os.path.abspath(
        os.path.join(os.path.dirname(__file__), '..', path)
    )


def determine_resource_absolute_file(path: str) -> Tuple[str, bool]:
    """
    Get and verify absolute path to resource file
    :param path: Relative path to resource
    :return: Tuple of absolute path to resource file and flag defining if is an external resource
    """
    # is resource exists as it is defined?
    initial_definition = get_external_resource_path(path)
    if os.path.isfile(initial_definition):
        return initial_definition, True

    # is resource exists as internal resource?
    patched_definition = get_internal_resource_path(path)
    if os.path.isfile(patched_definition):
        return patched_definition, False

    raise Exception('Requested resource %s is not exists at %s or %s' % (path, initial_definition, patched_definition))


def determine_resource_absolute_dir(path: str) -> Tuple[str, bool]:
    """
    Get and verify absolute path to resource directory
    :param path: Relative path to resource
    :return: Tuple of absolute path to resource directory and flag defining if is an external resource
    """
    dirname = os.path.dirname(path)
    # is resource dir exists as it is defined?
    initial_definition = get_external_resource_path(dirname)
    if os.path.isdir(initial_definition):
        return initial_definition, True

    # is resource dir exists as internal resource?
    patched_definition = get_internal_resource_path(dirname)
    if os.path.isdir(patched_definition):
        return patched_definition, False

    raise Exception(
        'Requested resource directory %s is not exists at %s or %s' % (path, initial_definition, patched_definition))


def load_yaml(filepath) -> dict:
    try:
        with open_utf8(filepath, 'r') as stream:
            return yaml.safe_load(stream)
    except yaml.YAMLError as exc:
        do_fail(f"Failed to load {filepath}", exc)


def true_or_false(value):
    """
    The method check string and boolean value
    :param value: Value that should be checked
    """
    input_string = str(value)
    if input_string in ['true', 'True', 'TRUE']:
        result = "true"
    elif input_string in ['false', 'False', 'FALSE']:
        result = "false"
    else:
        result = "undefined"
    return result


def get_version_filepath():
    return get_internal_resource_path("version")


def get_version():
    return read_internal(get_version_filepath()).strip()


class ClusterStorage:
    """
    File preservation:
    1- Create folder where dumps are stored
    2- Rotating dumps in the storage folder
    3- Uploading dumps to nodes
    4- Copying dumps to new nodes
    """

    PRESERVED_DUMP_FILES = ['procedure.yaml', 'procedure_parameters', 'cluster_precompiled.yaml',
                            'cluster.yaml','cluster_initial.yaml', 'cluster_finalized.yaml']

    def __init__(self, cluster):
        from kubemarine.core.cluster import KubernetesCluster
        self.cluster: KubernetesCluster = cluster
        self.dir_path = "/etc/kubemarine/procedures/"
        self.dir_name = ''
        self.dir_location = ''

    def make_dir(self):
        """
        This method creates a directory in which logs about operations on the cluster will be stored.
        """
        readable_timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        initial_procedure = self.cluster.context["initial_procedure"]
        self.dir_name = readable_timestamp + "_" + initial_procedure + "/"
        self.dir_location = self.dir_path + self.dir_name
        self.cluster.nodes['control-plane'].get_final_nodes().sudo(f"mkdir -p {self.dir_location} ; sudo rm {self.dir_path + 'latest_dump'} ;"
                                                 f" sudo ln -s {self.dir_location} {self.dir_path + 'latest_dump'}")

    def rotation_file(self):
        """
        This method packs files with logs and maintains a structured storage of logs on the cluster.
        """
        not_pack_file = self.cluster.inventory['procedure_history']['archive_threshold']
        delete_old = self.cluster.inventory['procedure_history']['delete_threshold']

        command = f'ls {self.dir_path} | grep -v latest_dump'
        node_group_results = self.cluster.nodes["control-plane"].get_final_nodes().sudo(command)
        with RemoteExecutor(self.cluster):
            for cxn, result in node_group_results.items():
                control_plane = self.cluster.make_group([cxn.host])
                files = result.stdout.split()
                files.sort(reverse=True)
                for i, file in enumerate(files):
                    if i >= not_pack_file and i < delete_old:
                        if 'tar.gz' not in file:
                            control_plane.sudo(f'tar -czvf {self.dir_path + file + ".tar.gz"} {self.dir_path + file} &&'
                                       f'sudo rm -r {self.dir_path + file}')
                    elif i >= delete_old:
                        control_plane.sudo(f'rm -rf {self.dir_path + file}')

    def compress_and_upload_archive(self):
        """
        This method compose dump files and sends the collected files to the nodes.
        """
        context = self.cluster.context
        archive = get_dump_filepath(context, "local.tar.gz")
        with tarfile.open(archive, "w:gz") as tar:
            for name in ClusterStorage.PRESERVED_DUMP_FILES:
                source = get_dump_filepath(context, name)
                if os.path.exists(source):
                    tar.add(source, 'dump/' + name)
            tar.add(context['execution_arguments']['config'], 'cluster.yaml')
            tar.add(get_version_filepath(), 'version')

        self.cluster.log.debug('Uploading archive with preserved information about the procedure.')
        self.cluster.nodes['control-plane'].get_final_nodes().put(archive, self.dir_location + 'local.tar.gz', sudo=True)
        self.cluster.nodes['control-plane'].get_final_nodes().sudo(f'tar -C {self.dir_location} -xzv --no-same-owner -f {self.dir_location + "local.tar.gz"}  && '
                                                 f'sudo rm -f {self.dir_location + "local.tar.gz"} ')

    def collect_procedure_info(self):
        """
        This method collects information about the type of procedure and the version of the tool we are working with.
        """
        context = self.cluster.context
        out = dict()
        out['arguments'] = context['initial_cli_arguments']
        if 'proceeded_tasks' in context:
            out['finished_tasks'] = context['proceeded_tasks']
        out["initial_procedure"] = context["initial_procedure"]
        out["successfully_performed"] = context["successfully_performed"]
        out['status'] = context['status']
        output = yaml.dump(out)
        dump_file(context, output, "procedure_parameters")

    def upload_info_new_control_planes(self):
        """
        This method is used to transfer backup logs from the initial control-plane to the new control-planes.
        """
        new_control_planes = self.cluster.nodes['control-plane'].get_new_nodes()
        if new_control_planes.is_empty():
            return

        node = self.cluster.nodes['control-plane'].get_initial_nodes().get_first_member(provide_node_configs=True)
        control_plane = self.cluster.make_group([node['connect_to']])
        data_copy_res = control_plane.sudo(f'tar -czvf /tmp/kubemarine-backup.tar.gz {self.dir_path}')
        self.cluster.log.verbose('Backup created:\n%s' % data_copy_res)
        control_plane.get('/tmp/kubemarine-backup.tar.gz',
                          get_dump_filepath(self.cluster.context, "dump_log_cluster.tar.gz"), 'dump_log_cluster.tar.gz')

        self.cluster.log.debug('Backup downloaded')

        for new_node in new_control_planes.get_ordered_members_list(provide_node_configs=True):
            group = self.cluster.make_group([new_node['connect_to']])
            group.put(get_dump_filepath(self.cluster.context, "dump_log_cluster.tar.gz"),
                      "/tmp/dump_log_cluster.tar.gz", sudo=True)
            group.sudo(f'tar -C / -xzvf /tmp/dump_log_cluster.tar.gz')



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import kubemarine.patches
from kubemarine.core import flow


def main(cli_arguments=None):
    cli_help = '''
    Script for automated update of the environment for the current version of Kubemarine.

    How to use:

    '''

    parser = flow.new_common_parser(cli_help)
    parser.add_argument('--force-skip', dest='skip', metavar='SKIP',
                        default='',
                        help='define comma-separated patches to skip')

    parser.add_argument('--force-apply', dest='apply', metavar='APPLY',
                        default='',
                        help='define explicit comma-separated set of patches to apply')

    parser.add_argument('--list',
                        action='store_true',
                        help='list all patches')

    parser.add_argument('--describe', metavar='PATCH',
                        help='describe the specified patch')

    context = flow.create_context(parser, cli_arguments, procedure="migrate_kubemarine")
    args = context['execution_arguments']

    patches = kubemarine.patches.patches
    patch_ids = [patch.identifier for patch in patches]

    if args['list']:
        if patch_ids:
            print("Available patches list:")
            for patch_id in patch_ids:
                print(patch_id)
        else:
            print("No patches available.")
        exit(0)

    if args['describe']:
        for patch in patches:
            if patch.identifier == args['describe']:
                print(patch.description)
                exit(0)
        print(f"Unknown patch '{args['describe']}'")
        exit(1)

    skip = [] if not args['skip'] else args['skip'].split(",")
    apply = [] if not args['apply'] else args['apply'].split(",")

    if apply and (set(apply) - set(patch_ids)):
        print(f"Unknown patches {list(set(apply) - set(patch_ids))}")
        exit(1)

    if skip and (set(skip) - set(patch_ids)):
        print(f"Unknown patches {list(set(skip) - set(patch_ids))}")
        exit(1)

    if apply:
        positions = [patch_ids.index(apply_id) for apply_id in apply]
        if not all(positions[i] < positions[i + 1] for i in range(len(positions) - 1)):
            print("Incorrect order of patches to apply. See --list for correct order of patches.")
            exit(1)

    actions = []
    for patch in patches:
        if apply:
            if patch.identifier in apply:
                actions.append(patch.action)
        elif skip:
            if patch.identifier not in skip:
                actions.append(patch.action)
        else:
            actions.append(patch.action)

    if not actions:
        print("No patches to apply")
        exit(0)
    flow.run_actions(context, actions)


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from io import StringIO

from kubemarine import system, packages
from kubemarine.core import utils
from kubemarine.core.executor import RemoteExecutor
from kubemarine.core.group import NodeGroup


def install(group: NodeGroup):
    with RemoteExecutor(group.cluster) as exe:
        for node in group.get_ordered_members_list(provide_node_configs=True):
            os_specific_associations = group.cluster.get_associations_for_node(node['connect_to'], 'docker')
            packages.install(node['connection'], include=os_specific_associations['package_name'])
            enable(node['connection'])

            # remove previous daemon.json to avoid problems in case when previous config was broken
            node['connection'].sudo("rm -f %s && sudo systemctl restart %s"
                                % (os_specific_associations['config_location'],
                                   os_specific_associations['service_name']))
    return exe.get_last_results_str()


def uninstall(group):
    # delete all known docker packages
    return packages.remove(group, include=['docker', 'docker-engine', 'docker.io', 'docker-ce'])


def enable(group: NodeGroup):
    # currently it is invoked only for single node
    service_name = group.cluster.get_package_association_for_node(group.get_host(), 'docker', 'service_name')
    system.enable_service(group, name=service_name, now=True)


def disable(group: NodeGroup):
    service_name = group.cluster.get_package_association_for_node(group.get_host(), 'docker', 'service_name')
    system.disable_service(group, name=service_name, now=True)


def configure(group: NodeGroup):
    log = group.cluster.log

    settings_json = json.dumps(group.cluster.inventory["services"]['cri']['dockerConfig'], sort_keys=True, indent=4)
    utils.dump_file(group.cluster, settings_json, 'docker-daemon.json')

    with RemoteExecutor(group.cluster) as exe:
        for node in group.get_ordered_members_list(provide_node_configs=True):
            os_specific_associations = group.cluster.get_associations_for_node(node['connect_to'], 'docker')
            log.debug("Uploading docker configuration to %s node..." % node['name'])
            node['connection'].put(StringIO(settings_json), os_specific_associations['config_location'], backup=True,
                                   sudo=True)
            log.debug("Restarting Docker on %s node..." % node['name'])
            node['connection'].sudo(f"chmod 600 {os_specific_associations['config_location']} && "
                                    f"sudo systemctl restart {os_specific_associations['service_name']} && "
                                    f"sudo {os_specific_associations['executable_name']} info")

    return exe.get_last_results_str()


def prune(group):
    return group.sudo('docker container stop $(sudo docker container ls -aq); '
                      'sudo docker container rm $(sudo docker container ls -aq); '
                      'sudo docker system prune -a -f; '
                      # kill all containerd-shim processes, so that no orphan containers remain 
                      'sudo pkill -9 -f "^containerd-shim"', warn=True, hide=True)



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import datetime
import io
import json
import os
import random
import shutil
import tarfile
import time
from collections import OrderedDict
import yaml

from kubemarine.core import utils, flow
from kubemarine.core.action import Action
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup
from kubemarine.core.resources import DynamicResources


def get_default_backup_files_list(cluster: KubernetesCluster):
    haproxy_service = cluster.get_package_association('haproxy', 'service_name')
    keepalived_service = cluster.get_package_association('keepalived', 'service_name')

    backup_files_list = [
        "/etc/resolv.conf",
        "/etc/hosts",
        "/etc/chrony.conf",
        "/etc/selinux/config",
        "/etc/yum.repos.d/",
        "/var/lib/kubelet/pki",
        "/etc/modules-load.d/",
        "/etc/audit/rules.d/",
        "/etc/haproxy/haproxy.cfg",
        f"/etc/systemd/system/{haproxy_service}.service.d/{haproxy_service}.conf",
        "/etc/keepalived/keepalived.conf",
        f"/etc/systemd/system/{keepalived_service}.service.d/{keepalived_service}.conf",
        "/usr/local/bin/check_haproxy.sh",
        "/etc/kubernetes",
        "/etc/systemd/system/kubelet.service"
    ]

    cri_impl = cluster.inventory['services']['cri']['containerRuntime']
    if cri_impl == "docker":
        backup_files_list.append("/etc/docker/daemon.json")
    elif cri_impl == "containerd":
        backup_files_list.append("/etc/containerd/config.toml")
        backup_files_list.append("/etc/crictl.yaml")

    return backup_files_list


def prepare_backup_tmpdir(cluster):
    backup_directory = cluster.context.get('backup_tmpdir')
    if not backup_directory:
        cluster.log.verbose('Backup directory is not ready yet, preparing..')
        backup_directory = cluster.context['backup_tmpdir'] = utils.get_dump_filepath(cluster.context, 'backup')
        shutil.rmtree(backup_directory, ignore_errors=True)
        os.mkdir(backup_directory)
        cluster.log.verbose('Backup directory prepared')
    return backup_directory


def verify_backup_location(cluster):
    target = utils.get_external_resource_path(cluster.procedure_inventory.get('backup_location', 'backup.tar.gz'))
    if not os.path.isdir(target) and not os.path.isdir(os.path.abspath(os.path.join(target, os.pardir))):
        raise FileNotFoundError('Backup location directory not exists')


def export_ansible_inventory(cluster):
    backup_directory = prepare_backup_tmpdir(cluster)
    shutil.copyfile(cluster.context['execution_arguments']['ansible_inventory_location'],
                    os.path.join(backup_directory, 'ansible-inventory.ini'))
    cluster.log.verbose('ansible-inventory.ini exported to backup')


def export_packages_list(cluster: KubernetesCluster):
    cluster.context['backup_descriptor']['nodes']['packages'] = {}
    if cluster.get_os_family() in ['rhel', 'rhel8']:
        cmd = r"rpm -qa"
    else:
        cmd = r"dpkg-query -f '${Package}=${Version}\n' -W"
    results = cluster.nodes['all'].sudo(cmd)
    for conn, result in results.items():
        cluster.context['backup_descriptor']['nodes']['packages'][conn.host] = result.stdout.strip().split('\n')


def export_hostname(cluster):
    cluster.context['backup_descriptor']['nodes']['hostnames'] = {}
    results = cluster.nodes['all'].sudo('hostnamectl status | head -n 1 | sed -e \'s/[a-zA-Z ]*://g\'')
    cluster.log.verbose(results)
    for conn, result in results.items():
        cluster.context['backup_descriptor']['nodes']['hostnames'][conn.host] = result.stdout.strip()


def export_cluster_yaml(cluster):
    backup_directory = prepare_backup_tmpdir(cluster)
    shutil.copyfile(utils.get_dump_filepath(cluster.context, 'cluster.yaml'),
                    os.path.join(backup_directory, 'cluster.yaml'))
    shutil.copyfile(utils.get_external_resource_path(cluster.context['execution_arguments']['config']),
                    os.path.join(backup_directory, 'original_cluster.yaml'))
    cluster.log.verbose('cluster.yaml exported to backup')


def export_nodes(cluster):
    backup_directory = prepare_backup_tmpdir(cluster)
    backup_nodes_data_dir = os.path.join(backup_directory, 'nodes_data')
    os.mkdir(backup_nodes_data_dir)

    backup_list = get_default_backup_files_list(cluster)
    for filepath, enabled in cluster.procedure_inventory.get('backup_plan', {}).get('nodes', {}).items():
        if not enabled and filepath in backup_list:
            backup_list.remove(filepath)
        if enabled and filepath not in backup_list:
            backup_list.append(filepath)

    cluster.log.debug('Backing up the following files: \n' + '  - ' + '\n  - '.join(backup_list))

    backup_command = 'cd /tmp && ' \
                     'sudo tar -czvf /tmp/kubemarine-backup.tar.gz -P $(sudo readlink -e %s) && ' \
                     'sudo ls -la /tmp/kubemarine-backup.tar.gz && ' \
                     'sudo du -hs /tmp/kubemarine-backup.tar.gz' % (' '.join(backup_list))

    data_copy_res = cluster.nodes['all'].run(backup_command)

    cluster.log.debug('Backup created:\n%s' % data_copy_res)

    cluster.log.debug('Downloading nodes backups:')
    for node in cluster.nodes['all'].get_ordered_members_list(provide_node_configs=True):
        node['connection'].get('/tmp/kubemarine-backup.tar.gz',
                               os.path.join(backup_nodes_data_dir, '%s.tar.gz' % node['name']))
        cluster.log.debug('Backup \'%s\' downloaded' % node['name'])

    cluster.log.verbose('Deleting backup file from nodes...')
    cluster.nodes['all'].sudo('rm -f /tmp/kubemarine-backup.tar.gz')


def export_etcd(cluster: KubernetesCluster):
    backup_directory = prepare_backup_tmpdir(cluster)
    etcd_node, is_custom_etcd_node = select_etcd_node(cluster)
    cluster.context['backup_descriptor']['etcd']['image'] = retrieve_etcd_image(cluster, etcd_node)

    # Try to detect cluster health and other metadata like db size, leader
    etcd_status = None
    try:
        result = etcd_node.sudo('etcdctl endpoint status --cluster -w json').get_simple_out()
        cluster.log.verbose(result)

        etcd_status = json.load(io.StringIO(result.lower()))
        parsed_etcd_status = {}
        for item in etcd_status:
            # get rid of https:// and :2379
            address = item['endpoint'].split('//')[1].split(':')[0]
            node_name = cluster.nodes['all'].get_first_member(apply_filter={"internal_address": address}, provide_node_configs=True)['name']
            parsed_etcd_status[node_name] = item
        cluster.context['backup_descriptor']['etcd']['status'] = parsed_etcd_status
    except Exception:
        cluster.log.verbose('Failed to load and parse ETCD status')

    if is_custom_etcd_node:
        cluster.context['backup_descriptor']['etcd']['source'] = etcd_node.get_nodes_names()[0]
    else:
        # if user did not provide node, then we have to select leader by ourselves
        if not etcd_status:
            raise Exception('Failed to load ETCD status and impossible to detect ETCD leader for making snapshot from it')
        etcd_leader_id = etcd_status[0]['status']['leader']
        etcd_leader_name = None
        for name, item in parsed_etcd_status.items():
            if item['status']['header']['member_id'] == etcd_leader_id:
                etcd_leader_name = name
        if etcd_leader_name:
            cluster.log.verbose('Detected ETCD leader: %s' % etcd_leader_name)
            cluster.context['backup_descriptor']['etcd']['source'] = etcd_leader_name
            etcd_node = cluster.nodes['control-plane'].get_member_by_name(etcd_leader_name)
        else:
            raise Exception('Failed to detect ETCD leader - not possible to create backup from actual DB')

    snap_name = 'snapshot%s.db' % int(round(time.time() * 1000))
    endpoint_ip = etcd_node.get_ordered_members_list(provide_node_configs=True)[0]["internal_address"]
    cluster.log.debug('Creating ETCD backup "%s"...' % snap_name)
    result = etcd_node.sudo(f'etcdctl snapshot save /var/lib/etcd/{snap_name} --endpoints=https://{endpoint_ip}:2379 '
                            f'&& sudo mv /var/lib/etcd/{snap_name} /tmp/{snap_name} '
                            f'&& sudo ls -la /tmp/{snap_name} '
                            f'&& sudo du -hs /tmp/{snap_name} '
                            f'&& sudo chmod 666 /tmp/{snap_name}', timeout=600)
    cluster.log.debug(result)
    etcd_node.get('/tmp/' + snap_name, backup_directory + '/etcd.db')
    cluster.log.verbose('Deleting ETCD snapshot file from "%s"...')
    etcd_node.sudo('rm -f /tmp/%s' % snap_name)


def select_etcd_node(cluster):
    custom_etcd_node = cluster.procedure_inventory.get('backup_plan', {}).get('etcd', {}).get('source_node')

    if custom_etcd_node:
        etcd_node = cluster.nodes['all'].get_member_by_name(custom_etcd_node)
        if etcd_node is None:
            raise Exception('Unknown ETCD node selected as source')
        return etcd_node, True
    else:
        return cluster.nodes['control-plane'].get_any_member(), False


def retrieve_etcd_image(cluster, etcd_node):
    # TODO: Detect ETCD version via /etc/kubernetes/manifests/etcd.yaml config if presented, otherwise use containers
    node_name = etcd_node.get_nodes_names()[0]
    if "docker" == cluster.inventory['services']['cri']['containerRuntime']:
        cont_inspect = "docker inspect $(sudo docker ps -a | grep etcd-%s | awk '{print $1; exit}')" % node_name
        etcd_container_json = json.loads(list(etcd_node.sudo(cont_inspect).values())[0].stdout)[0]
        etcd_image_sha = etcd_container_json['Image'][7:]  # remove "sha256:" prefix

        images_result = etcd_node.sudo("docker image ls --format '{{json .}}'")
        formatted_images_result = "[" + ",".join(list(images_result.values())[0].stdout.strip().split('\n')) + "]"
        images_json = json.loads(formatted_images_result)
        for image in images_json:
            if image['ID'] == etcd_image_sha[:len(image['ID'])]:
                return f"{image['Repository']}:{image['Tag']}"
    else:
        cont_search = "sudo crictl ps --label io.kubernetes.pod.name=etcd-%s -aq | awk '{print $1; exit}'" % node_name
        cont_inspect = f"crictl inspect $({cont_search})"
        etcd_container_json = json.loads(list(etcd_node.sudo(cont_inspect).values())[0].stdout)
        etcd_image_sha = etcd_container_json['info']['config']['image']['image']

        images_json = json.loads(list(etcd_node.sudo("crictl images -v -o json").values())[0].stdout)['images']
        for image in images_json:
            if image['id'] == etcd_image_sha:
                return image['repoTags'][0]

    raise Exception("Unable to find etcd image on node %s" % node_name)


def export_kubernetes_version(cluster: KubernetesCluster):
    control_plane = cluster.nodes['control-plane'].get_any_member()
    version = control_plane.sudo('kubectl get nodes --no-headers | head -n 1 | awk \'{print $5; exit}\'').get_simple_out()
    cluster.context['backup_descriptor']['kubernetes']['version'] = version.strip()


# There is no way to parallel resources connection via Queue or Pool:
# the ssh connection is not possible to parallelize due to thread lock
def download_resources(log, resources, location, control_plane: NodeGroup, namespace=None):

    if namespace:
        log.debug('Downloading resources from namespace "%s"...' % namespace)

    actual_resources = []

    if not resources:
        log.debug('No resources found to download')
        return actual_resources

    cmd = ''
    resource_separator = ''.join(random.choice('=-_') for _ in range(32))

    for resource in resources:
        if cmd != '':
            cmd += ' && echo \'\n' + resource_separator + '\n\' && '
        if namespace:
            cmd += 'sudo kubectl -n %s get --ignore-not-found %s -o yaml' % (namespace, resource)
        else:
            cmd += 'sudo kubectl get --ignore-not-found %s -o yaml' % resource

    result = control_plane.sudo(cmd).get_simple_out()
    control_plane.cluster.log.verbose(result)

    found_resources_results = result.split(resource_separator)
    for i, result in enumerate(found_resources_results):
        resource = resources[i]
        resource_file_path = os.path.join(location, '%s.yaml' % resource)
        result = result.strip()
        if result and result != '':
            actual_resources.append(resource)
            with utils.open_external(resource_file_path, 'w') as resource_file_stream:
                resource_file_stream.write(result)

    return actual_resources


def export_kubernetes(cluster):
    backup_directory = prepare_backup_tmpdir(cluster)
    control_plane = cluster.nodes['control-plane'].get_any_member()

    cluster.log.debug('Loading namespaces:')
    namespaces_result = control_plane.sudo('kubectl get ns -o yaml')
    cluster.log.verbose(namespaces_result)
    namespaces_string = list(namespaces_result.values())[0].stdout.strip()
    namespaces_yaml = yaml.safe_load(namespaces_string)

    proposed_namespaces = cluster.procedure_inventory.get('backup_plan', {}).get('kubernetes', {}).get('namespaced_resources', {}).get('namespaces', 'all')
    namespaces = []
    for item in namespaces_yaml['items']:
        name = item['metadata']['name']
        if proposed_namespaces == 'all' or name in proposed_namespaces:
            cluster.log.verbose('Namespace "%s" added' % name)
            namespaces.append(name)
        else:
            cluster.log.verbose('Namespace "%s" excluded' % name)

    if proposed_namespaces != 'all':
        for proposed_namespace in proposed_namespaces:
            if proposed_namespace not in namespaces:
                raise Exception('Proposed namespace "%s" not found in loaded cluster namespaces' % proposed_namespace)

    cluster.log.debug(namespaces)
    kubernetes_res_dir = os.path.join(backup_directory, 'kubernetes_resources')
    os.mkdir(kubernetes_res_dir)

    cluster.log.debug('Loading namespaced resources:')
    resources_result = control_plane.sudo('kubectl api-resources --verbs=list --namespaced -o name '
                                   '| grep -v "events.events.k8s.io" | grep -v "events" | sort | uniq')
    cluster.log.verbose(resources_result)
    parsed_resources = list(resources_result.values())[0].stdout.strip().split('\n')
    proposed_resources = cluster.procedure_inventory.get('backup_plan', {}).get('kubernetes', {}).get('namespaced_resources', {}).get('resources', 'all')

    resources = [resource for resource in parsed_resources if proposed_resources == 'all' or resource in proposed_resources]

    for resource in parsed_resources:
        if resource not in resources:
            cluster.log.verbose('Resource "%s" excluded' % resource)

    if proposed_resources != 'all':
        for proposed_resource in proposed_resources:
            if proposed_resource not in resources:
                raise Exception('Proposed resource "%s" not found in loaded cluster resources' % proposed_resource)

    cluster.log.debug(resources)

    namespaced_resources_map = {}
    total_files = 0

    for namespace in namespaces:
        namespace_dir = os.path.join(kubernetes_res_dir, namespace)
        os.mkdir(namespace_dir)
        actual_resources = download_resources(cluster.log, resources, namespace_dir, control_plane, namespace)
        if actual_resources:
            total_files += len(actual_resources)
            namespaced_resources_map[namespace] = actual_resources

    cluster.log.debug('Loading non-namespaced resources:')
    resources_result = control_plane.sudo('kubectl api-resources --verbs=list --namespaced=false -o name '
                                   '| grep -v "events.events.k8s.io" | grep -v "events" | sort | uniq')
    cluster.log.verbose(resources_result)
    parsed_resources = list(resources_result.values())[0].stdout.strip().split('\n')
    proposed_resources = cluster.procedure_inventory.get('backup_plan', {}).get('kubernetes', {}).get('nonnamespaced_resources', 'all')

    resources = [resource for resource in parsed_resources if proposed_resources == 'all' or resource in proposed_resources]

    for resource in parsed_resources:
        if resource not in resources:
            cluster.log.verbose('Resource "%s" excluded' % resource)

    if proposed_resources != 'all':
        for proposed_resource in proposed_resources:
            if proposed_resource not in resources:
                raise Exception('Proposed resource "%s" not found in loaded cluster resources' % proposed_resource)

    cluster.log.debug(resources)

    nonnamespaced_resources_list = download_resources(cluster.log, resources, kubernetes_res_dir, control_plane)
    total_files += len(nonnamespaced_resources_list)

    cluster.log.verbose('Total files saved: %s' % total_files)

    if namespaced_resources_map:
        cluster.context['backup_descriptor']['kubernetes']['resources']['namespaced'] = namespaced_resources_map

    if nonnamespaced_resources_list:
        cluster.context['backup_descriptor']['kubernetes']['resources']['nonnamespaced'] = nonnamespaced_resources_list


def make_descriptor(cluster):
    backup_directory = prepare_backup_tmpdir(cluster)

    cluster.context['backup_descriptor']['kubernetes']['thirdparties'] = cluster.inventory['services']['thirdparties']
    cluster.context['backup_descriptor']['meta']['time']['finished'] = datetime.datetime.now()

    with utils.open_external(os.path.join(backup_directory, 'descriptor.yaml'), 'w') as output:
        output.write(yaml.dump(cluster.context['backup_descriptor']))


def pack_data(cluster):
    cluster_name = cluster.inventory['cluster_name']
    backup_directory = prepare_backup_tmpdir(cluster)

    backup_filename = 'backup-%s-%s.tar.gz' % (cluster_name, utils.get_current_timestamp_formatted())

    target = utils.get_external_resource_path(cluster.procedure_inventory.get('backup_location', backup_filename))
    if os.path.isdir(target):
        target = os.path.join(target, backup_filename)

    cluster.log.debug('Packing all data...')
    with tarfile.open(target, "w:gz") as tar_handle:
        for root, dirs, files in os.walk(backup_directory):
            for file in files:
                pathname = os.path.join(root, file)
                tar_handle.add(pathname, pathname.replace(backup_directory, ''))
        tar_handle.close()

    cluster.log.verbose('Cleaning up...')
    shutil.rmtree(backup_directory, ignore_errors=True)


tasks = OrderedDict({
    "verify_backup_location": verify_backup_location,
    "export": {
        "inventory": {
            "cluster_yaml": export_cluster_yaml,
            "ansible_inventory": export_ansible_inventory,
        },
        "lists": {
            "rpms": export_packages_list,
            "hostname": export_hostname,
        },
        "nodes": export_nodes,
        "etcd": export_etcd,
        "cluster_version": export_kubernetes_version,
        "kubernetes": export_kubernetes,
    },
    "make_descriptor": make_descriptor,
    "pack": pack_data,
})


class BackupAction(Action):
    def __init__(self):
        super().__init__('backup')

    def run(self, res: DynamicResources):
        flow.run_tasks(res, tasks)


def main(cli_arguments=None):
    cli_help = '''
    Script for making backup of Kubernetes resources and nodes contents.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks)

    context = flow.create_context(parser, cli_arguments, procedure='backup')
    context['execution_arguments']['disable_dump'] = False
    context['backup_descriptor'] = {
        'meta': {
            'time': {
                'started': datetime.datetime.now()
            }
        },
        'etcd': {},
        'nodes': {},
        'kubernetes': {
            'resources': {}
        }
    }

    flow.run_actions(context, [BackupAction()])


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
from copy import deepcopy

from kubemarine.core import utils, static
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroupResult, NodeGroup


def enrich_inventory_apply_upgrade_defaults(inventory, cluster):
    if cluster.context.get('initial_procedure') == 'upgrade':
        upgrade_version = cluster.context["upgrade_version"]
        upgrade_thirdparties = cluster.procedure_inventory.get(upgrade_version, {}).get('thirdparties')
        if upgrade_thirdparties:
            upgrade_thirdparties = deepcopy(upgrade_thirdparties)
            default_thirdparties = static.DEFAULTS['services']['thirdparties']

            # keep some configurations (unpack) from default thirdparties, if they are not re-defined
            for destination, config in upgrade_thirdparties.items():
                if destination in default_thirdparties and 'unpack' in default_thirdparties[destination]\
                        and 'unpack' not in config:
                    config['unpack'] = deepcopy(default_thirdparties[destination]['unpack'])

            inventory['services']['thirdparties'] = upgrade_thirdparties
        else:
            cluster.log.warning('New thirdparties for upgrade procedure is not set in procedure config - default will be used')
    return inventory


def get_thirdparty_recommended_sha(destination, cluster):
    cluster.log.verbose("Calculate recommended sha for thirdparty %s..." % destination)
    # Get kubeadm version
    kubernetes_version = cluster.inventory['services']['kubeadm']['kubernetesVersion']
    effective_kubernetes_version = ".".join(kubernetes_version.split('.')[0:2])

    # Get software versions map
    software_name = cluster.globals['thirdparties'].get(destination, {}).get('software_name', None)
    software_versions = cluster.globals['compatibility_map']['software'].get(software_name, {})

    # Return sha1 related to used kubernetes version
    recommended_sha = software_versions[kubernetes_version].get('sha1', None) \
        if kubernetes_version in software_versions else \
        software_versions.get(effective_kubernetes_version, {}).get('sha1', None)

    if recommended_sha is not None:
        cluster.log.verbose(f"Recommended sha for thirdparty {destination} was calculated: {recommended_sha}")
    else:
        cluster.log.verbose(f"Recommended sha for thirdparty {destination} doesn't exist")

    return recommended_sha


def enrich_inventory_apply_defaults(inventory, cluster):
    # if thirdparties is empty, then nothing to do
    if not inventory['services'].get('thirdparties', {}):
        return inventory
    raw_inventory = cluster.raw_inventory

    for destination, config in inventory['services']['thirdparties'].items():

        if isinstance(config, str):
            config = {
                'source': config
            }

        if config.get('mode') is None:
            config['mode'] = 700

        if config.get('owner') is None:
            config['owner'] = 'root'

        if config.get('group') is not None:
            config['groups'] = [config['group']]
            del config['group']

        if config.get('node') is not None:
            config['nodes'] = [config['node']]
            del config['node']

        if config.get('groups') is None and config.get('nodes') is None:
            config['groups'] = ['control-plane', 'worker']

        if config.get('nodes') is not None:
            all_nodes_names = cluster.nodes['all'].get_nodes_names()
            for node_name in config['nodes']:
                if node_name not in all_nodes_names:
                    raise Exception('Unknown node name provided for thirdparty %s. '
                                    'Expected any of %s, but \'%s\' found.'
                                    % (destination, all_nodes_names, node_name))

        raw_config = raw_inventory.get('services', {}).get('thirdparties', {}).get(destination, {})
        recommended_sha = get_thirdparty_recommended_sha(destination, cluster)
        if 'source' not in raw_config and 'sha1' not in raw_config and recommended_sha is not None:
            config['sha1'] = get_thirdparty_recommended_sha(destination, cluster)

        inventory['services']['thirdparties'][destination] = config

    # remove "crictl" from thirdparties when docker is used, but ONLY IF it is NOT explicitly specified in cluster.yaml
    cri_name = inventory['services']['cri']['containerRuntime']
    crictl_key = '/usr/bin/crictl.tar.gz'
    if cri_name == "docker" and \
            crictl_key not in cluster.raw_inventory.get('services', {}).get('thirdparties', {}) and \
            crictl_key in inventory['services']['thirdparties']:
        del(inventory['services']['thirdparties'][crictl_key])

    return inventory


def get_install_group(cluster: KubernetesCluster, config: dict):
    return cluster.create_group_from_groups_nodes_names(
        config.get('groups', []), config.get('nodes', []))


def get_group_require_unzip(cluster: KubernetesCluster, inventory: dict) -> NodeGroup:
    thirdparties: dict = inventory['services']['thirdparties']

    group = cluster.make_group([])
    for destination, config in thirdparties.items():
        extension = destination.split('.')[-1]
        if config.get('unpack') is None or extension != 'zip':
            continue

        install_group = get_install_group(cluster, config)
        group = group.include_group(install_group)

    return group


def install_thirdparty(filter_group: NodeGroup, destination: str) -> NodeGroupResult or None:
    cluster = filter_group.cluster
    config = cluster.inventory['services'].get('thirdparties', {}).get(destination)

    if config is None:
        raise Exception('Not possible to install thirdparty %s - not found in configfile' % destination)

    common_group = get_install_group(cluster, config)
    common_group = common_group.intersection_group(filter_group)

    if common_group.is_empty():
        cluster.log.verbose(f'No destination nodes to install thirdparty {destination!r}')
        return

    cluster.log.debug("Thirdparty \"%s\" will be installed" % destination)
    is_curl = config['source'][:4] == 'http' and '://' in config['source'][4:8]

    # all commands will be grouped to single run
    remote_commands = ''

    # directory will be created if it is not exists
    destination_directory = '/'.join(destination.split('/')[:-1])
    cluster.log.verbose('Destination directory: %s' % destination_directory)

    # ! In the further code there is no error and nothing is missing !
    # Here a long shell command is intentionally constructed and executed at once to speed up work
    # At the same time, in the middle of the construction of the command, a file may suddenly be uploaded and then
    # the command will be executed in two runs instead of single run

    # is destination directory exists?
    remote_commands += 'mkdir -p %s' % destination_directory

    if is_curl:
        cluster.log.verbose('Installation via curl download detected')
        if config.get('sha1') is not None:
            cluster.log.verbose('SHA1 hash is defined, it will be used during installation')
            # if hash equal, then stop further actions immediately! unpack should not be performed too
            remote_commands += ' && FILE_HASH=$(sudo openssl sha1 %s | sed "s/^.* //"); ' \
                               '[ "%s" == "${FILE_HASH}" ] && exit 0 || true ' % (destination, config['sha1'])
        remote_commands += ' && sudo rm -f %s && sudo curl --max-time %d -f -g -L %s -o %s && ' % (destination, cluster.inventory['timeout_download'], config['source'], destination)
    else:
        cluster.log.verbose('Installation via sftp upload detected')
        cluster.log.debug(common_group.sudo(remote_commands))
        remote_commands = ''
        # TODO: Possible use SHA1 from inventory instead of calculating if provided?
        script = utils.read_internal(config['source'])
        common_group.put(io.StringIO(script), destination, sudo=True)

        # TODO: Do not upload local files if they already exists on remote machines

    remote_commands += 'sudo chmod %s %s' % (config['mode'], destination)
    remote_commands += ' && sudo chown %s %s' % (config['owner'], destination)
    remote_commands += ' && sudo ls -la %s' % destination

    if config.get('unpack') is not None:
        cluster.log.verbose('Unpack request detected')

        remote_commands += ' && sudo mkdir -p %s' % config['unpack']

        extension = destination.split('.')[-1]
        if extension == 'zip':
            cluster.log.verbose('Unzip will be used for unpacking')
            # TODO re-installation waits forever
            remote_commands += ' && sudo unzip %s -d %s' % (destination, config['unpack'])
        else:
            cluster.log.verbose('Tar will be used for unpacking')
            remote_commands += ' && sudo tar -zxf %s -C %s' % (destination, config['unpack'])

        # TODO access rights do not work for zip
        remote_commands += ' && sudo tar -tf %s | xargs -I FILE sudo chmod %s %s/FILE' \
                           % (destination, config['mode'], config['unpack'])
        remote_commands += ' && sudo tar -tf %s | xargs -I FILE sudo chown %s %s/FILE' \
                           % (destination, config['owner'], config['unpack'])
        remote_commands += ' && sudo tar -tf %s | xargs -I FILE sudo ls -la %s/FILE' % (destination, config['unpack'])

    return common_group.sudo(remote_commands)


def install_all_thirparties(group):
    cluster = group.cluster
    log = cluster.log

    if not cluster.inventory['services'].get('thirdparties', {}):
        return

    for destination in cluster.inventory['services']['thirdparties'].keys():
        skip_thirdparty = False

        if cluster.context.get("initial_procedure") != "add_node":
            # TODO: speed up algorithm via else/continue/break
            for plugin_name, plugin_configs in cluster.inventory['plugins'].items():
                for plugin_procedure in plugin_configs['installation']['procedures']:
                    if plugin_procedure.get('thirdparty') == destination:
                        log.verbose('Thirdparty \'%s\' should be installed with \'%s\' plugin'
                                    % (destination, plugin_name))
                        skip_thirdparty = True

        if skip_thirdparty:
            log.verbose('Thirdparty %s installation delayed' % destination)
        else:
            res = install_thirdparty(group, destination)
            if res is not None:
                log.debug(res)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import time
from typing import Tuple, List, Dict, Callable, Any, Optional
from contextvars import Token, ContextVar

import fabric
import invoke

from fabric.connection import Connection
from concurrent.futures.thread import ThreadPoolExecutor

GRE = ContextVar('KubemarineGlobalRemoteExecutor', default=None)


class RemoteExecutor:

    def __init__(self, cluster,
                 warn=False,
                 lazy=True,
                 parallel=True,
                 ignore_failed=False,
                 enforce_children=False,
                 timeout=None):
        self.cluster = cluster
        # TODO revise necessity of warn option
        self.warn = False
        self.lazy = lazy
        self.parallel = parallel
        # TODO support ignore_failed option.
        #  Probably it should be chosen automatically depending on warn=? of commands kwargs (not of the same executor option).
        self.ignore_failed = False
        # TODO revise this option
        self.enforce_children = False
        self.timeout = timeout
        self.connections_queue: Dict[Connection, List[Tuple]] = {}
        self._last_token = -1
        self.previous_context_token = Token.MISSING
        self.command_separator = ''.join(random.choice('=-_') for _ in range(32))
        self.results = []
        self.connections_queue_history = []

    def __del__(self):
        pass

    def __enter__(self):
        executor = self._get_active_executor()
        if executor == self or not executor.enforce_children:
            self.previous_context_token = GRE.set(self)
        return executor

    def __exit__(self, exc_type, exc_value, tb):
        if self.previous_context_token != Token.MISSING:
            GRE.reset(self.previous_context_token)
        if self.connections_queue:
            self.flush()
            # Throw any exceptions ONLY when no handlers waits for results. For non-queued results any exceptions should
            # handled via _do() or other parent error handling mechanism. When queue flushed inside RemoteExecutor
            # context block, then nobody can handle results, so RemoteExecutor should throw an error by itself.
            # TODO: Merge results/exceptions handling with kubemarine.core.group.NodeGroup._do_with_wa to avoid code dup
            if not self.warn:
                self.throw_on_failed()

    def _get_active_executor(self):
        executor = GRE.get()
        if executor:
            return executor
        else:
            return self

    def _is_actions_equal(self, action1, action2):
        if action1[0] not in ["sudo", "run"] or action1[0] != action2[0]:
            return False
        # todo bug, what if one kwargs dict is a subdict of another?
        for key, value in action1[2].items():
            if value != action2[2].get(key):
                return False
        return True

    def reparse_results(self, results, batch):
        batch_no_cnx: Dict[str, tuple] = {}
        conns_by_host: Dict[str, Connection] = {}
        for cnx, data in batch.items():
            batch_no_cnx[cnx.host] = data
            conns_by_host[cnx.host] = cnx
        executor = self._get_active_executor()
        reparsed_results = {}
        for host, result in results.items():
            conn_results = {}
            action, callbacks, tokens = batch_no_cnx[host]
            if isinstance(result,
                          fabric.runners.Result) and executor.command_separator in result.stdout and executor.command_separator in result.stderr:
                stderrs = result.stderr.split(executor.command_separator + '\n')
                raw_stdouts = result.stdout.split(executor.command_separator + '\n')
                stdouts = []
                exit_codes = []
                i = 0
                while i < len(raw_stdouts):
                    stdouts.append(raw_stdouts[i])
                    if i + 1 < len(raw_stdouts):
                        exit_codes.append(int(raw_stdouts[i + 1].strip()))
                    i += 2
                exit_codes.append(result.exited)
                for i, code in enumerate(exit_codes):
                    token = tokens[i]
                    conn_results[token] = fabric.runners.Result(stdout=stdouts[i], stderr=stderrs[i], exited=code,
                                                                connection=conns_by_host[host])
            else:
                conn_results[tokens[0]] = result
            reparsed_results[conns_by_host[host]] = conn_results
        # TODO: In long term run, collect callbacks and wait for them
        return reparsed_results

    def _merge_actions(self, actions):
        executor = self._get_active_executor()

        if executor.ignore_failed:
            # todo exit codes in separators do not work, because 'echo command_separator' always rewrites exit code to 0
            separator_symbol = ";"
        else:
            separator_symbol = "&&"

        separator = f" {separator_symbol} " \
                    f"echo \"{executor.command_separator}\" {separator_symbol} " \
                    f"echo $? {separator_symbol} " \
                    f"echo \"{executor.command_separator}\" {separator_symbol} " \
                    f"echo \"{executor.command_separator}\" 1>&2 {separator_symbol} "

        merged_actions = []

        for payload in actions:
            action, callback, token = payload
            if merged_actions and executor._is_actions_equal(merged_actions[-1][0], action):
                precommand = ''
                if action[0] == 'sudo':
                    precommand = 'sudo '
                previous_action = merged_actions[-1][0]
                merged_action_command = previous_action[1][0] + separator + precommand + action[1][0]
                merged_actions[-1][0] = (previous_action[0], tuple([merged_action_command]), previous_action[2])
                merged_actions[-1][1].append(callback)
                merged_actions[-1][2].append(token)
            else:
                merged_actions.append([action, [callback], [token]])

        return merged_actions

    def _get_callables(self):
        executor = self._get_active_executor()
        callables = {}

        for connection, actions in executor.connections_queue.items():
            callables[connection] = executor._merge_actions(actions)

        i = 0
        batches = []

        while i != -1:
            batch = {}
            for conn, actions in callables.items():
                if len(actions) > i:
                    batch[conn] = actions[i]
            if not batch:
                i = -1
            else:
                i += 1
                batches.append(batch)

        return batches

    def queue(self, target, action: Tuple, callback: Callable = None) -> int or dict:
        # TODO support callbacks
        callback = None
        executor = self._get_active_executor()
        executor._last_token = token = executor._last_token + 1

        if isinstance(target, Connection):
            target = [Connection]
        if isinstance(target, dict):
            target = list(target.values())

        if not target:
            executor.cluster.log.verbose('Connections list is empty, nothing to queue')
        else:
            for connection in target:
                if not executor.connections_queue.get(connection):
                    executor.connections_queue[connection] = []
                executor.connections_queue[connection].append((action, callback, token))

        if not executor.lazy:
            return executor.flush()
        else:
            return token

    def reset_queue(self) -> None:
        executor = self._get_active_executor()
        executor.connections_queue = {}

    def get_last_results(self):
        executor = self._get_active_executor()
        if len(executor.results) == 0:
            return None
        return executor.results[-1]

    def get_merged_nodegroup_results(self, filter_tokens: Optional[List[int]] = None):
        """
        Merges last tokenized results into NodeGroupResult.

        The method is useful to check exceptions, to check result in case only one command per node is executed,
        or to check result of specific commands in the batch if filter_tokens parameter is provided.

        :param filter_tokens: tokens to filter result of specific commands in the batch
        :return: None or NodeGroupResult
        """
        # TODO: Get rid of this WA import, added to avoid circular import problem
        from kubemarine.core.group import NodeGroupResult
        executor = self._get_active_executor()
        if len(executor.results) == 0:
            return None
        group_results = NodeGroupResult(self.cluster)
        for cxn, host_results in executor.get_last_results().items():
            merged_result = {
                "stdout": '',
                "stderr": '',
                "exited": 0
            }
            object_result = None

            tokens = host_results.keys() if filter_tokens is None else filter_tokens
            for token, result in host_results.items():
                if isinstance(result, Exception):
                    object_result = result
                    continue
                elif token not in tokens:
                    continue

                if isinstance(result, fabric.runners.Result):
                    if result.stdout:
                        merged_result['stdout'] += result.stdout
                    if result.stderr:
                        merged_result['stderr'] += result.stderr

                    # Exit codes can not be merged, that's why they are assigned by priority:
                    # 1. Most important code is 1, it should be assigned if any results contains it
                    # 2. Non-zero exit code from last command
                    # 3. Zero exit code, when all commands succeeded
                    if result.exited == 1 or (
                            merged_result['exited'] != 1 and result.exited != merged_result['exited']):
                        merged_result['exited'] = result.exited
                else:
                    object_result = result

            # Some commands can produce non-parsed objects, like 'timeout'
            # In that case it is impossible to merge something, and last such an "object" should be passed as a result
            if object_result is not None:
                group_results[cxn] = object_result
            else:
                group_results[cxn] = fabric.runners.Result(stdout=merged_result['stdout'],
                                                           stderr=merged_result['stderr'],
                                                           exited=merged_result['exited'],
                                                           connection=cxn)
        return group_results

    def throw_on_failed(self):
        """
        Throws an exception if last results has failed commands. Ignores enabled executor.warn option.
        :return: None
        """
        self.get_merged_result()

    def get_merged_result(self):
        """
        Returns last results, merged into NodeGroupResult. If any node failed, throws GroupException.

        The method is useful to propagate exceptions, or to check result in case only one command per node is executed.

        :return: NodeGroupResult
        """
        executor = self._get_active_executor()
        if len(executor.results) == 0:
            return None
        group_results = executor.get_merged_nodegroup_results()
        if len(group_results.failed) > 0:
            # If any failed, then throw exception.
            # Do not check 'warn' arg, because if any command had 'False' value, UnexpectedExit can be already thrown.
            raise fabric.group.GroupException(group_results)

        return group_results

    def get_last_results_str(self):
        batched_results = self.get_last_results()
        if not batched_results:
            return
        output = ""
        for conn, results in batched_results.items():
            for token, result in results.items():
                if isinstance(result, invoke.exceptions.UnexpectedExit):
                    result = result.result

                # for now we do not know how-to print transfer result
                if not isinstance(result, fabric.runners.Result):
                    continue

                if output != "":
                    output += "\n"
                output += "\t%s (%s): code=%i" % (conn.host, token, result.exited)
                if result.stdout:
                    output += "\n\t\tSTDOUT: %s" % result.stdout.replace("\n", "\n\t\t        ")
                if result.stderr:
                    output += "\n\t\tSTDERR: %s" % result.stderr.replace("\n", "\n\t\t        ")

        return output

    def print_last_results(self) -> None:
        """
        Prints debug message to log with last executed queued command results.
        :return: None
        """
        results = self.get_last_results_str()
        self.cluster.log.debug(results)

    def flush(self) -> dict:
        executor = self._get_active_executor()

        batch_results = {}

        if not executor.connections_queue:
            executor.cluster.log.verbose('Queue is empty, nothing to perform')
            return batch_results

        callable_batches = executor._get_callables()

        max_workers = len(executor.connections_queue.keys())
        if not executor.parallel:
            max_workers = 1

        with ThreadPoolExecutor(max_workers=max_workers) as TPE:
            for batch in callable_batches:
                results = {}
                futures = {}

                def safe_exec(result_map: Dict[str, Any], key: str, call: Callable[[], Any]):
                    try:
                        # sleep required to avoid thread starvation
                        time.sleep(0.1)
                        result_map[key] = call()
                        time.sleep(0.1)
                    except Exception as e:
                        results[key] = e

                # todo if previous batch for some cxn failed, probably we should not continue executing for that cxn
                for cxn, payload in batch.items():
                    action, callbacks, tokens = payload
                    do_type, args, kwargs = action
                    executor.cluster.log.verbose('Executing %s %s with options: %s' % (do_type, args, kwargs))
                    safe_exec(futures, cxn.host, lambda: TPE.submit(getattr(cxn, do_type), *args, **kwargs))

                for host, future in futures.items():
                    safe_exec(results, host, lambda: future.result(timeout=executor.timeout))

                parsed_results = executor.reparse_results(results, batch)
                for host, tokenized_results in parsed_results.items():
                    if not batch_results.get(host):
                        batch_results[host] = {}
                    for token, res in tokenized_results.items():
                        batch_results[host][token] = res

        executor.connections_queue_history.append(executor.connections_queue)
        executor.reset_queue()
        executor.results.append(batch_results)

        return batch_results



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from importlib import import_module
from copy import deepcopy
from typing import Optional

import yaml

from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.errors import KME
from kubemarine import jinja
from kubemarine.core import utils, static
from kubemarine.core.yaml_merger import default_merger

# All enrichment procedures should not connect to any node.
# The information about nodes should be collected within KubernetesCluster#detect_nodes_context().
DEFAULT_ENRICHMENT_FNS = [
    "kubemarine.core.schema.verify_inventory",
    "kubemarine.core.defaults.merge_defaults",
    "kubemarine.kubernetes.add_node_enrichment",
    "kubemarine.kubernetes.remove_node_enrichment",
    "kubemarine.controlplane.controlplane_node_enrichment",
    "kubemarine.core.defaults.append_controlplain",
    "kubemarine.kubernetes.enrich_upgrade_inventory",
    "kubemarine.plugins.enrich_upgrade_inventory",
    "kubemarine.packages.enrich_inventory_associations",
    "kubemarine.packages.enrich_inventory_packages",
    "kubemarine.system.enrich_upgrade_inventory",
    "kubemarine.core.defaults.compile_inventory",
    "kubemarine.core.defaults.manage_true_false_values",
    "kubemarine.admission.manage_enrichment",
    "kubemarine.thirdparties.enrich_inventory_apply_upgrade_defaults",
    "kubemarine.procedures.migrate_cri.enrich_inventory",
    "kubemarine.core.defaults.apply_registry",
    "kubemarine.core.defaults.calculate_node_names",
    "kubemarine.core.defaults.verify_node_names",
    "kubemarine.core.defaults.apply_defaults",
    "kubemarine.keepalived.enrich_inventory_apply_defaults",
    "kubemarine.haproxy.enrich_inventory",
    "kubemarine.kubernetes.enrich_inventory",
    "kubemarine.admission.enrich_inventory",
    "kubemarine.kubernetes_accounts.enrich_inventory",
    "kubemarine.plugins.calico.enrich_inventory",
    "kubemarine.plugins.nginx_ingress.cert_renew_enrichment",
    "kubemarine.plugins.nginx_ingress.enrich_inventory",
    "kubemarine.core.defaults.calculate_nodegroups",
    "kubemarine.keepalived.enrich_inventory_calculate_nodegroup",
    "kubemarine.thirdparties.enrich_inventory_apply_defaults",
    "kubemarine.system.verify_inventory",
    "kubemarine.system.enrich_etc_hosts",
    "kubemarine.packages.enrich_inventory_include_all",
    "kubemarine.audit.verify_inventory",
    "kubemarine.plugins.enrich_inventory",
    "kubemarine.plugins.verify_inventory",
    "kubemarine.coredns.enrich_add_hosts_config",
    "kubemarine.k8s_certs.renew_verify",
    "kubemarine.cri.enrich_inventory"
]

supported_defaults = {
    'rbac': {
        'account_defaults': 'accounts'
    },
    'node_defaults': 'nodes',
    'plugin_defaults': 'plugins',
}

invalid_node_name_regex = re.compile("[^a-z-.\\d]", re.M)
escaped_expression_regex = re.compile('({%[\\s*|]raw[\\s*|]%}.*?{%[\\s*|]endraw[\\s*|]%})', re.M)
jinja_query_regex = re.compile("{{ .* }}", re.M)


def apply_defaults(inventory, cluster: KubernetesCluster):
    recursive_apply_defaults(supported_defaults, inventory)

    for i, node in enumerate(inventory["nodes"]):
        address = cluster.get_access_address_from_node(node)

        # we have definitely know how to connect
        cluster.inventory["nodes"][i]["connect_to"] = address
        cluster.inventory["nodes"][i]["connection"] = cluster.make_group([address])

        if not cluster.context["nodes"].get(address):
            cluster.context["nodes"][address] = {}

        if address not in cluster.ips["all"]:
            cluster.ips['all'].append(address)

        for role in node.get("roles"):
            if role not in cluster.roles:
                cluster.roles.append(role)
                cluster.ips[role] = []
            if address not in cluster.ips[role]:
                cluster.ips[role].append(address)

    return inventory


def apply_registry(inventory, cluster):

    if not inventory.get('registry'):
        cluster.log.verbose('Unified registry is not used')
        return inventory

    thirdparties_address = None
    containerd_endpoints = None
    protocol = None

    # registry contains either 'endpoints' or 'address' that is validated by JSON schema.
    if inventory['registry'].get('endpoints'):
        registry_mirror_address, containerd_endpoints, thirdparties_address = apply_registry_endpoints(inventory, cluster)
    else:
        if inventory['registry'].get('docker_port'):
            registry_mirror_address = "%s:%s" % (inventory['registry']['address'], inventory['registry']['docker_port'])
        else:
            registry_mirror_address = inventory['registry']['address']

        protocol = 'http'
        if inventory['registry'].get('ssl', False):
            protocol = 'https'

        if inventory['registry'].get('webserver', False):
            thirdparties_address = f"{protocol}://{inventory['registry']['address']}"

    # Patch kubeadm imageRepository
    if not inventory['services']['kubeadm'].get('imageRepository'):
        inventory['services']['kubeadm']["imageRepository"] = registry_mirror_address

    # it is necessary to convert URIs from quay.io/xxx:v1 to example.com:XXXX/xxx:v1
    if inventory.get('plugin_defaults') is None:
        inventory['plugin_defaults'] = {}
    if inventory['plugin_defaults'].get('installation') is None:
        inventory['plugin_defaults']['installation'] = {}
    if inventory['plugin_defaults']['installation'].get('registry') is None:
        inventory['plugin_defaults']['installation']['registry'] = registry_mirror_address

    # The following section rewrites DEFAULT plugins registries and do not touches user-defined registries in plugins
    # This section required, because plugins defaults contains default non-docker registries and method
    # "kubemarine.core.defaults.recursive_apply_defaults" will not overwrite this default registries, because it can not
    # distinguish default from user-defined.
    # Also, this part of code supports plugin_defaults inventory section and applies everything in accordance with the
    # priority of the registries.
    for plugin_name, plugin_params in cluster.inventory['plugins'].items():
        if cluster.inventory['plugins'][plugin_name].get('installation') is None:
            cluster.inventory['plugins'][plugin_name]['installation'] = {}
        if cluster.raw_inventory.get('plugins', {}).get(plugin_name, {}).get('installation', {}).get('registry') is None:
            cluster.inventory['plugins'][plugin_name]['installation']['registry'] = inventory['plugin_defaults']['installation']['registry']

    cri_impl = inventory['services']['cri']['containerRuntime']

    if cri_impl == "docker" and inventory['registry'].get('endpoints'):
        raise KME('KME0007')

    if cri_impl == "docker":

        if protocol == 'http':
            if inventory['services']['cri']['dockerConfig'].get("insecure-registries") is None:
                inventory['services']['cri']['dockerConfig']["insecure-registries"] = []
            insecure_registries = inventory['services']['cri']['dockerConfig']["insecure-registries"]
            insecure_registries.append(registry_mirror_address)
            inventory['services']['cri']['dockerConfig']["insecure-registries"] = list(set(insecure_registries))

        if inventory['services']['cri']['dockerConfig'].get("registry-mirrors") is None:
            inventory['services']['cri']['dockerConfig']["registry-mirrors"] = []

        registry_mirrors = inventory['services']['cri']['dockerConfig']["registry-mirrors"]
        registry_mirrors.append(f"{protocol}://{registry_mirror_address}")
        inventory['services']['cri']['dockerConfig']["registry-mirrors"] = list(set(registry_mirrors))

    elif cri_impl == "containerd":
        registry_section = f'plugins."io.containerd.grpc.v1.cri".registry.mirrors."{registry_mirror_address}"'

        if not inventory['services']['cri']['containerdConfig'].get(registry_section):
            if not containerd_endpoints:
                containerd_endpoints = ["%s://%s" % (protocol, registry_mirror_address)]

            inventory['services']['cri']['containerdConfig'][registry_section] = {
                'endpoint': containerd_endpoints
            }

        effective_kubernetes_version = ".".join(inventory['services']['kubeadm']['kubernetesVersion'].split('.')[0:2])
        pause_version = cluster.globals['compatibility_map']['software']['pause'][effective_kubernetes_version]['version']
        if not inventory['services']['cri']['containerdConfig'].get('plugins."io.containerd.grpc.v1.cri"'):
            inventory['services']['cri']['containerdConfig']['plugins."io.containerd.grpc.v1.cri"'] = {}
        if not inventory['services']['cri']['containerdConfig']['plugins."io.containerd.grpc.v1.cri"'].get('sandbox_image'):
            inventory['services']['cri']['containerdConfig']['plugins."io.containerd.grpc.v1.cri"']['sandbox_image'] = \
                f"{inventory['services']['kubeadm']['imageRepository']}/pause:{pause_version}"

    if inventory['services'].get('thirdparties', []) and thirdparties_address:
        for destination, config in inventory['services']['thirdparties'].items():

            if isinstance(config, str):
                new_source = inventory['services']['thirdparties'][destination]
            elif config.get('source') is not None:
                new_source = inventory['services']['thirdparties'][destination]['source']
            else:
                continue

            for binary in ['kubeadm', 'kubelet', 'kubectl']:
                if destination == '/usr/bin/' + binary:
                    new_source = new_source.replace('https://storage.googleapis.com/kubernetes-release/release',
                                                    '%s/kubernetes/%s'
                                                    % (thirdparties_address, binary))

            if '/usr/bin/calicoctl' == destination:
                new_source = new_source.replace('https://github.com/projectcalico/calico/releases/download',
                                                '%s/projectcalico/calico'
                                                % thirdparties_address)

            if '/usr/bin/crictl.tar.gz' == destination:
                new_source = new_source.replace('https://github.com/kubernetes-sigs/cri-tools/releases/download',
                                                '%s/kubernetes-sigs/cri-tools'
                                                % thirdparties_address)
            if isinstance(config, str):
                inventory['services']['thirdparties'][destination] = new_source
            else:
                inventory['services']['thirdparties'][destination]['source'] = new_source

    return inventory


def apply_registry_endpoints(inventory, cluster):

    if not inventory['registry'].get('mirror_registry'):
        inventory['registry']['mirror_registry'] = 'registry.cluster.local'

    registry_mirror_address = inventory['registry']['mirror_registry']

    # todo Currently registry.endpoints is used only for containerd registry mirrors, but it can be provided explicitly.
    #  Probably we could make endpoints optional in this case.
    containerd_endpoints = inventory['registry']['endpoints']
    thirdparties_address = inventory['registry'].get('thirdparties')

    return registry_mirror_address, containerd_endpoints, thirdparties_address


def append_controlplain(inventory, cluster: Optional[KubernetesCluster]):

    if inventory.get('control_plain', {}).get('internal') and inventory.get('control_plain', {}).get('external'):
        if cluster:
            cluster.log.verbose('Control plains are set manually, nothing to detect.')
        return inventory

    if cluster:
        cluster.log.verbose('Detecting control plains...')

    # calculate controlplain ips
    internal_address = None
    internal_address_source = None
    external_address = None
    external_address_source = None

    # vrrp_ip section is not enriched yet
    # todo what if ip is an ip of some node to remove?
    if inventory.get('vrrp_ips'):
        for i, item in enumerate(inventory['vrrp_ips']):
            if isinstance(item, str):
                if internal_address is None:
                    internal_address = item
                    internal_address_source = 'vrrp_ip[%s]' % i
            # todo remove p1_migrate_not_bind_vrrp_fix after next release
            elif item.get('params', {}).get('maintenance-type', False) != 'not bind' \
                    or (cluster and not cluster.context.get('p1_migrate_not_bind_vrrp_fix', True)):
                if internal_address is None or item.get('control_endpoint', False):
                    internal_address = item['ip']
                    internal_address_source = 'vrrp_ip[%s]' % i
                if item.get('floating_ip') and (external_address is None or item.get('control_endpoint', False)):
                    external_address = item['floating_ip']
                    external_address_source = 'vrrp_ip[%s]' % i

    if internal_address is not None and external_address is None and cluster:
        cluster.log.warning('VRRP_IPs has an internal address, but do not have an external one. Your configuration may be incorrect. Trying to handle this problem automatically...')

    if internal_address is None or external_address is None:
        # 'master' role is not deleted due to unit tests are not refactored
        for role in ['balancer', 'master', 'control-plane']:
            # nodes are not compiled to groups yet
            for node in inventory['nodes']:
                if role in node['roles'] and 'remove_node' not in node['roles']:
                    if internal_address is None or node.get('control_endpoint', False):
                        internal_address = node['internal_address']
                        internal_address_source = role
                        if node.get('name'):
                            internal_address_source += ' \"%s\"' % node['name']
                    if node.get('address') and (external_address is None or node.get('control_endpoint', False)):
                        external_address = node['address']
                        external_address_source = role
                        if node.get('name'):
                            external_address_source += ' \"%s\"' % node['name']

    if external_address is None:
        cluster.log.warning('Failed to detect external control plain. Something may work incorrect!')
        external_address = internal_address

    if cluster:
        cluster.log.debug('Control plains:\n   Internal: %s (%s)\n   External: %s (%s)' % (internal_address, internal_address_source, external_address, external_address_source))

    # apply controlplain ips
    if not inventory.get('control_plain'):
        inventory['control_plain'] = {}

    if not inventory['control_plain'].get('internal'):
        inventory['control_plain']['internal'] = internal_address

    if not inventory['control_plain'].get('external'):
        inventory['control_plain']['external'] = external_address

    return inventory


def recursive_apply_defaults(defaults, section):
    for key, value in defaults.items():
        if isinstance(value, dict) and section.get(key) is not None and section[key]:
            recursive_apply_defaults(value, section[key])
        # check if target section exists and not empty
        elif section.get(value) is not None and section[value]:

            if isinstance(section[value], list):
                for i, v in enumerate(section[value]):
                    # copy defaults as new dict, to avoid problems with memory links
                    node_config = deepcopy(section[key])

                    # update defaults with custom-defined node configs
                    # TODO: Use deepmerge instead of update
                    node_config.update(v)

                    # replace old node config with merged one
                    section[value][i] = node_config

            else:
                # deepcopy the whole section, otherwise it will break dict while replacing
                section_copy = deepcopy(section[value])
                for custom_key, custom_value in section_copy.items():
                    # here section['key'] refers to default, not custom value
                    default_value = deepcopy(section[key])
                    section[value][custom_key] = default_merger.merge(default_value, custom_value)


def calculate_node_names(inventory: dict, cluster):
    roles_iterators = {}
    for i, node in enumerate(inventory['nodes']):
        for role_name in ['control-plane', 'worker', 'balancer']:
            if role_name in node['roles']:
                # The idea is this:
                # If the name is already specified, we must skip this node,
                # however, we must consider that we already have a node of this type
                # and increase this type iterator
                # As a result, we get such an algorithm. For example, with the following inventory:
                #
                # - name: k8s-control-plane-1, roles: ['control-plane']
                # - roles: ['control-plane']
                # - name: k8s-control-plane-3, roles: ['control-plane']
                #
                # We should get the following calculation result:
                #
                # - name: k8s-control-plane-1, roles: ['control-plane']
                # - name: control-plane-2, roles: ['control-plane']
                # - name: k8s-control-plane-3, roles: ['control-plane']
                #
                role_i = roles_iterators.get(role_name, 1)
                roles_iterators[role_name] = role_i + 1
                if node.get('name') is None:
                    inventory['nodes'][i]['name'] = '%s-%s' % (role_name, role_i)
    return inventory


def verify_node_names(inventory, cluster):
    known_names = []
    for i, node in enumerate(inventory['nodes']):
        node_name = node['name']
        if node_name in known_names:
            raise Exception('Node name %s is duplicated in configfile' % node_name)
        if re.findall(invalid_node_name_regex, node_name):
            raise Exception('Node name \"%s\" contains invalid characters. A DNS-1123 subdomain must consist of lower '
                            'case alphanumeric characters, \'-\' or \'.\'' % node_name)
        known_names.append(node_name)
    return inventory


def calculate_nodegroups(inventory, cluster):
    for role in cluster.ips.keys():
        cluster.nodes[role] = cluster.make_group(cluster.ips[role])
    return inventory


def merge_defaults(inventory: dict, cluster: KubernetesCluster):
    base_inventory = deepcopy(static.DEFAULTS)

    inventory = default_merger.merge(base_inventory, inventory)
    # it is necessary to temporary put half-compiled inventory to cluster inventory field
    cluster._inventory = inventory
    return inventory


def enrich_inventory(cluster: KubernetesCluster, inventory: dict, make_dumps=True, enrichment_functions=None):
    if not enrichment_functions:
        enrichment_functions = DEFAULT_ENRICHMENT_FNS

    # run required fields calculation
    for enrichment_fn in enrichment_functions:
        fn_package_name, fn_method_name = enrichment_fn.rsplit('.', 1)
        mod = import_module(fn_package_name)
        cluster.log.verbose('Calling fn "%s"' % enrichment_fn)
        inventory = getattr(mod, fn_method_name)(inventory, cluster)

    cluster.log.verbose('Enrichment finished!')

    if make_dumps:
        from kubemarine import controlplane
        inventory_for_dump = controlplane.controlplane_finalize_inventory(cluster, prepare_for_dump(inventory))
        utils.dump_file(cluster, yaml.dump(inventory_for_dump, ), "cluster.yaml")

    return inventory


def compile_inventory(inventory, cluster):

    # convert references in yaml to normal values
    iterations = 100
    root = deepcopy(inventory)
    root['globals'] = static.GLOBALS

    while iterations > 0:

        cluster.log.verbose('Inventory is not rendered yet...')
        inventory = compile_object(cluster.log, inventory, root)

        temp_dump = yaml.dump(inventory)

        # remove golang specific
        temp_dump = re.sub(escaped_expression_regex, '', temp_dump.replace('\n', ''))

        # it is necessary to carry out several iterations,
        # in case we have dynamic variables that reference each other
        if '{{' in temp_dump or '{%' in temp_dump:
            iterations -= 1
        else:
            iterations = 0

    inventory = compile_object(cluster.log, inventory, root, ignore_jinja_escapes=False)

    from kubemarine import controlplane
    inventory_for_dump = controlplane.controlplane_finalize_inventory(cluster, prepare_for_dump(inventory))
    merged_inventory = yaml.dump(inventory_for_dump)
    utils.dump_file(cluster, merged_inventory, "cluster_precompiled.yaml")

    return inventory


def compile_object(log, struct, root, ignore_jinja_escapes=True):
    if isinstance(struct, list):
        new_struct = []
        for i, v in enumerate(struct):
            struct[i] = compile_object(log, v, root, ignore_jinja_escapes=ignore_jinja_escapes)
            # delete empty list entries, which can appear after jinja compilation
            if struct[i] != '':
                new_struct.append(struct[i])
        struct = new_struct
    elif isinstance(struct, dict):
        for k, v in struct.items():
            struct[k] = compile_object(log, v, root, ignore_jinja_escapes=ignore_jinja_escapes)
    elif isinstance(struct, str) and ('{{' in struct or '{%' in struct):
        struct = compile_string(log, struct, root, ignore_jinja_escapes=ignore_jinja_escapes)
    return struct


def compile_string(log, struct, root, ignore_jinja_escapes=True):
    log.verbose("Rendering \"%s\"" % struct)

    if ignore_jinja_escapes:
        iterator = escaped_expression_regex.finditer(struct)
        struct = re.sub(escaped_expression_regex, '', struct)
        struct = jinja.new(log, root).from_string(struct).render(**root)

        for match in iterator:
            span = match.span()
            struct = struct[:span[0]] + match.group() + struct[span[0]:]
    else:
        struct = jinja.new(log, root).from_string(struct).render(**root)

    log.verbose("\tRendered as \"%s\"" % struct)
    return struct


def escape_jinja_characters_for_inventory(cluster: KubernetesCluster, obj):
    if isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = escape_jinja_characters_for_inventory(cluster, value)
    elif isinstance(obj, list):
        for key, value in enumerate(obj):
            obj[key] = escape_jinja_characters_for_inventory(cluster, value)
    elif isinstance(obj, str):
        obj = _escape_jinja_character(obj)
    return obj


def _escape_jinja_character(value):
    if '{{' in value and '}}' in value and re.search(jinja_query_regex, value):
        matches = re.findall(jinja_query_regex, value)
        for match in matches:
            # TODO: rewrite to correct way of match replacement: now it can cause "{raw}{raw}xxx.." circular bug
            value = value.replace(match, '{% raw %}'+match+'{% endraw %}')
    return value


def prepare_for_dump(inventory, copy=True):
    # preparation for dump required to remove memory links

    if copy:
        dump_inventory = deepcopy(inventory)
    else:
        dump_inventory = inventory

    for i, node in enumerate(dump_inventory['nodes']):
        if 'connection' in dump_inventory['nodes'][i]:
            del dump_inventory['nodes'][i]['connection']

    return dump_inventory


def manage_true_false_values(inventory, cluster):
    # Check undefined values for plugin.name.install and convert it to bool
    for plugin_name, plugin_item in inventory["plugins"].items():
        # Check install value
        if 'install' not in plugin_item:
            continue
        value = utils.true_or_false(plugin_item.get('install', False))
        if value == 'undefined':
            raise ValueError(f"Found unsupported value for plugin.{plugin_name}.install: {plugin_item['install']}")
        plugin_item['install'] = value == 'true'
    return inventory



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml

from kubemarine import system
from kubemarine.core import utils

import io


def enrich_add_hosts_config(inventory, cluster):
    if not inventory['services']['coredns']['configmap'].get('Hosts'):
        inventory['services']['coredns']['configmap']['Hosts'] = system.generate_etc_hosts_config(inventory, cluster)
    return inventory


def proceed_section_keyvalue(data, tabsize):
    tab = " "*tabsize
    config = ''

    for key, value in data.items():
        if isinstance(value, bool):
            if value:
                config += '\n' + tab + '%s' % key
            continue
        if isinstance(value, str) or isinstance(value, int) and value:
            if not isinstance(value, int) and any((c in set(' ')) for c in value):
                config += '\n' + tab + '%s \"%s\"' % (key, value)
            elif isinstance(value, str) and len(value) == 0:
                config += '\n' + tab + '%s' % key
            else:
                config += '\n' + tab + '%s %s' % (key, value)
            continue
        if isinstance(value, list) and value:
            config += '\n' + tab + '%s %s' % (key, " ".join(value))
            continue
        if isinstance(value, dict):
            config += generate_nested_sections(key, value, tabsize)
            continue
        raise Exception('Unknown type of field in coredns services')

    return config


def generate_nested_sections(type, data, tabsize):
    tab = " "*tabsize
    config = ''

    max_priority = 0
    for section_name, section_value in data.items():
        if section_value.get('priority') is not None and section_value['priority'] > max_priority:
            max_priority = section_value['priority']

    iterated = 0
    sections = []
    for section_name, section_value in data.items():
        if section_value.get('priority') is None:
            iterated += 1
            section_priority = max_priority + iterated
        else:
            section_priority = section_value['priority']

        if section_value.get('enabled', True) in ['1', 1, True, 'True']:
            sections.append({
                'name': section_name,
                'priority': section_priority
            })

    sections = sorted(sections, key=lambda i: i['priority'])

    for section in sections:

        if type == 'kubernetes':
            config += '\n' + tab + type
            if data[section['name']].get('zone'):
                if isinstance(data[section['name']]['zone'], list):
                    data[section['name']]['zone'] = ' '.join(data[section['name']]['zone'])
                config += ' ' + data[section['name']]['zone']
            config += ' {' + proceed_section_keyvalue(data[section['name']]['data'], tabsize + 2) + '\n' + tab + '}'

        elif type == 'hosts':
            config += '\n' + tab + type
            if data[section['name']].get('file') and isinstance(data[section['name']]['file'], str):
                config += ' ' + data[section['name']]['file']
            config += ' {' + proceed_section_keyvalue(data[section['name']]['data'], tabsize + 2) + '\n' + tab + '}'

        elif type == 'template':
            zones = [None]
            if data[section['name']].get('zone'):
                zones = data[section['name']]['zone']
                if isinstance(zones, str):
                    zones = [zones]
            for zone in zones:
                config += '\n' + tab + type
                if data[section['name']].get('class'):
                    config += ' ' + data[section['name']]['class']
                if data[section['name']].get('type'):
                    config += ' ' + data[section['name']]['type']
                if zone:
                    config += ' ' + zone
                config += ' {' + proceed_section_keyvalue(data[section['name']]['data'], tabsize + 2) + '\n' + tab + '}'

        else:
            config += '\n' + tab + type + ' {' + proceed_section_keyvalue(data[section['name']]['data'], tabsize + 2)\
                      + '\n' + tab + '}'

    return config


def generate_configmap(inventory):
    config = '''apiVersion: v1
kind: ConfigMap
metadata:
  name: coredns
  namespace: kube-system
data:'''

    for config_type, data in inventory['services']['coredns']['configmap'].items():
        config += '\n  %s: |' % config_type
        if config_type == 'Corefile':
            for port, settings in data.items():
                config += '\n    %s {' % port
                config += proceed_section_keyvalue(settings, 6)
                config += '\n    }'
        else:
            config += '\n    ' + data.replace('\n', '\n    ')
    return config + '\n'


def apply_configmap(cluster, config):
    utils.dump_file(cluster, config, 'coredns-configmap.yaml')

    group = cluster.nodes['control-plane'].include_group(cluster.nodes.get('worker')).get_final_nodes()
    group.put(io.StringIO(config), '/etc/kubernetes/coredns-configmap.yaml', backup=True, sudo=True)

    return cluster.nodes['control-plane'].get_final_nodes().get_first_member()\
        .sudo('kubectl apply -f /etc/kubernetes/coredns-configmap.yaml && '
             'sudo kubectl rollout restart -n kube-system deployment/coredns')


def apply_patch(cluster):
    apply_command = ''

    for config_type in ['deployment']:

        if not cluster.inventory['services']['coredns'].get(config_type):
            continue

        if apply_command != '':
            apply_command += ' && sudo '

        config = yaml.dump(cluster.inventory['services']['coredns'][config_type])
        filename = 'coredns-%s-patch.yaml' % config_type
        filepath = '/etc/kubernetes/' + filename

        utils.dump_file(cluster, config, filename)

        group = cluster.nodes['control-plane'].include_group(cluster.nodes.get('worker')).get_final_nodes()
        group.put(io.StringIO(config), filepath, backup=True, sudo=True)

        apply_command = 'kubectl patch %s coredns -n kube-system --type merge -p \"$(sudo cat %s)\"' % (config_type, filepath)

    if apply_command == '':
        return 'Nothing to patch'

    return cluster.nodes['control-plane'].get_final_nodes().get_first_member().sudo(apply_command)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from kubemarine.core import summary
from kubemarine.core.cluster import KubernetesCluster


def schedule_summary_report(cluster: KubernetesCluster):
    plugin_item = cluster.inventory['plugins']['kubernetes-dashboard']
    hostname = plugin_item['hostname']
    # Currently we declare that Dashboard UI is available only via HTTPS
    summary.schedule_report(cluster.context, summary.SummaryItem.DASHBOARD_URL, f'https://{hostname}')
